import importlib
import os
import pkgutil

import click
import yaml

from rich.console import Console
from rich.panel import Panel
from rich.style import Style
from rich.text import Text


def from_yaml(filename: str):
  try:
    with open(filename, 'r') as f:
      return yaml.safe_load(f)
  except yaml.YAMLError as e:
    click.echo(f"Error reading YAML file: {e}", err=True)
    return {}


def dump_yaml(data, filename: str):
  try:
    with open(filename, 'w') as f:
      yaml.dump(data, f)
  except Exception as e:
    click.echo(f"Error writing YAML file: {e}", err=True)


def set_base_url(env):
  environments = {
      'prod': 'https://api.clarifai.com',
      'staging': 'https://api-staging.clarifai.com',
      'dev': 'https://api-dev.clarifai.com'
  }

  if env in environments:
    return environments[env]
  else:
    raise ValueError("Invalid environment. Please choose from 'prod', 'staging', 'dev'.")


# Dynamically find and import all command modules from the cli directory
def load_command_modules():
  package_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cli')

  for _, module_name, _ in pkgutil.iter_modules([package_dir]):
    if module_name not in ['base', '__main__']:  # Skip the base.py and __main__ file itself
      importlib.import_module(f'clarifai.cli.{module_name}')


def display_co_resources(response, resource_type):
  """Display compute orchestration resources listing results using rich."""

  console = Console()
  panel = Panel(
      Text(f"List of {resource_type}s", justify="center"),
      title="",
      style=Style(color="blue", bold=True),
      border_style="green",
      width=60)
  console.print(panel)
  for indx, item in enumerate(list(response)):
    panel = Panel(
        "\n".join([f"{'ID'}: {item.id}", f"{'Description'}: {item.description}"]),
        title=f"{resource_type} {(indx + 1)}",
        border_style="green",
        width=60)
    console.print(panel)



import os

import click

from ..utils.cli import dump_yaml, from_yaml, load_command_modules, set_base_url


class CustomMultiGroup(click.Group):

  def group(self, *args, **kwargs):
    """Behaves the same as `click.Group.group()` except if passed
        a list of names, all after the first will be aliases for the first.
        """

    def decorator(f):
      aliased_group = []
      if isinstance(args[0], list):
        # we have a list so create group aliases
        _args = [args[0][0]] + list(args[1:])
        for alias in args[0][1:]:
          grp = super(CustomMultiGroup, self).group(alias, *args[1:], **kwargs)(f)
          grp.short_help = "Alias for '{}'".format(_args[0])
          aliased_group.append(grp)
      else:
        _args = args

      # create the main group
      grp = super(CustomMultiGroup, self).group(*_args, **kwargs)(f)

      # for all of the aliased groups, share the main group commands
      for aliased in aliased_group:
        aliased.commands = grp.commands

      return grp

    return decorator


@click.group(cls=CustomMultiGroup)
@click.pass_context
def cli(ctx):
  """Clarifai CLI"""
  ctx.ensure_object(dict)
  config_path = 'config.yaml'
  if os.path.exists(config_path):
    ctx.obj = from_yaml(config_path)
  else:
    ctx.obj = {}


@cli.command()
@click.option('--config', type=click.Path(), required=False, help='Path to the config file')
@click.option(
    '-e',
    '--env',
    required=False,
    help='Environment to use, choose from prod, staging and dev',
    type=click.Choice(['prod', 'staging', 'dev']))
@click.option('--user_id', required=False, help='User ID')
@click.pass_context
def login(ctx, config, env, user_id):
  """Login command to set PAT and other configurations."""

  if config and os.path.exists(config):
    ctx.obj = from_yaml(config)

  if 'pat' in ctx.obj:
    os.environ["CLARIFAI_PAT"] = ctx.obj['pat']
    click.echo("Loaded PAT from config file.")
  elif 'CLARIFAI_PAT' in os.environ:
    ctx.obj['pat'] = os.environ["CLARIFAI_PAT"]
    click.echo("Loaded PAT from environment variable.")
  else:
    _pat = click.prompt(
        "Get your PAT from https://clarifai.com/settings/security and pass it here", type=str)
    os.environ["CLARIFAI_PAT"] = _pat
    ctx.obj['pat'] = _pat
    click.echo("PAT saved successfully.")

  if user_id:
    ctx.obj['user_id'] = user_id
    os.environ["CLARIFAI_USER_ID"] = ctx.obj['user_id']
  elif 'user_id' in ctx.obj:
    ctx.obj['user_id'] = ctx.obj.get('user_id', "")
    os.environ["CLARIFAI_USER_ID"] = ctx.obj['user_id']
  elif 'CLARIFAI_USER_ID' in os.environ:
    ctx.obj['user_id'] = os.environ["CLARIFAI_USER_ID"]
  else:
    user_id = click.prompt("Pass the User ID here", type=str)
    os.environ["CLARIFAI_USER_ID"] = user_id
    ctx.obj['user_id'] = user_id
    click.echo("User ID saved successfully.")

  if env:
    ctx.obj['env'] = env
    ctx.obj['base_url'] = set_base_url(env)
    os.environ["CLARIFAI_API_BASE"] = ctx.obj['base_url']
  elif 'env' in ctx.obj:
    ctx.obj['env'] = ctx.obj.get('env', "prod")
    ctx.obj['base_url'] = set_base_url(ctx.obj['env'])
    os.environ["CLARIFAI_API_BASE"] = ctx.obj['base_url']
  elif 'CLARIFAI_API_BASE' in os.environ:
    ctx.obj['base_url'] = os.environ["CLARIFAI_API_BASE"]

  dump_yaml(ctx.obj, 'config.yaml')


# Import the CLI commands to register them
load_command_modules()


def main():
  cli()



import click
from clarifai.cli.base import cli
from clarifai.client.nodepool import Nodepool
from clarifai.utils.cli import display_co_resources, from_yaml


@cli.group(['deployment', 'dpl'])
def deployment():
  """Manage Deployments: create, delete, list"""
  pass


@deployment.command()
@click.option(
    '-np_id',
    '--nodepool_id',
    required=False,
    help='Nodepool ID for the Nodepool to interact with.')
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=True,
    help='Path to the deployment config file.')
@click.option(
    '-dpl_id',
    '--deployment_id',
    required=False,
    help='New deployment ID for the deployment to create.')
@click.pass_context
def create(ctx, nodepool_id, config, deployment_id):
  """Create a new Deployment with the given config file."""
  if not nodepool_id:
    deployment_config = from_yaml(config)
    nodepool_id = deployment_config['deployment']['nodepools'][0]['id']

  nodepool = Nodepool(
      nodepool_id=nodepool_id,
      user_id=ctx.obj['user_id'],
      pat=ctx.obj['pat'],
      base_url=ctx.obj['base_url'])
  if deployment_id:
    nodepool.create_deployment(config, deployment_id=deployment_id)
  else:
    nodepool.create_deployment(config)


@deployment.command()
@click.option(
    '-np_id',
    '--nodepool_id',
    required=True,
    help='Nodepool ID for the Nodepool to interact with.')
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=16)
@click.pass_context
def list(ctx, nodepool_id, page_no, per_page):
  """List all deployments for the nodepool."""

  nodepool = Nodepool(
      nodepool_id=nodepool_id,
      user_id=ctx.obj['user_id'],
      pat=ctx.obj['pat'],
      base_url=ctx.obj['base_url'])
  response = nodepool.list_deployments(page_no=page_no, per_page=per_page)
  display_co_resources(response, "Deployment")


@deployment.command()
@click.option(
    '-np_id',
    '--nodepool_id',
    required=True,
    help='Nodepool ID for the Nodepool to interact with.')
@click.option('-dpl_id', '--deployment_id', help='Deployment ID of the nodepool to delete.')
@click.pass_context
def delete(ctx, nodepool_id, deployment_id):
  """Deletes a deployment for the nodepool."""

  nodepool = Nodepool(
      nodepool_id=nodepool_id,
      user_id=ctx.obj['user_id'],
      pat=ctx.obj['pat'],
      base_url=ctx.obj['base_url'])
  nodepool.delete_deployments([deployment_id])



import importlib
import inspect
import os
import re
import sys
import tarfile
import time
from string import Template

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf import json_format
from rich import print
from rich.markup import escape

from clarifai.client import BaseClient
from clarifai.runners.models.model_class import ModelClass
from clarifai.runners.utils.const import (AVAILABLE_PYTHON_IMAGES, AVAILABLE_TORCH_IMAGES,
                                          CONCEPTS_REQUIRED_MODEL_TYPE, DEFAULT_PYTHON_VERSION,
                                          PYTHON_BASE_IMAGE, TORCH_BASE_IMAGE)
from clarifai.runners.utils.loader import HuggingFaceLoader
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import logger
from clarifai.versions import CLIENT_VERSION

# parse the user's requirements.txt to determine the proper base image to build on top of, based on the torch and other large dependencies and it's versions
# List of dependencies to look for
dependencies = [
    'torch',
    'clarifai',
]


def _clear_line(n: int = 1) -> None:
  LINE_UP = '\033[1A'  # Move cursor up one line
  LINE_CLEAR = '\x1b[2K'  # Clear the entire line
  for _ in range(n):
    print(LINE_UP, end=LINE_CLEAR, flush=True)


class ModelBuilder:
  DEFAULT_CHECKPOINT_SIZE = 50 * 1024**3  # 50 GiB

  def __init__(self, folder: str, validate_api_ids: bool = True, download_validation_only=False):
    """
    :param folder: The folder containing the model.py, config.yaml, requirements.txt and
    checkpoints.
    :param validate_api_ids: Whether to validate the user_id and app_id in the config file. TODO(zeiler):
    deprecate in favor of download_validation_only.
    :param download_validation_only: Whether to skip the API config validation. Set to True if
    just downloading a checkpoint.
    """
    self._client = None
    if not validate_api_ids:  # for backwards compatibility
      download_validation_only = True
    self.download_validation_only = download_validation_only
    self.folder = self._validate_folder(folder)
    self.config = self._load_config(os.path.join(self.folder, 'config.yaml'))
    self._validate_config()
    self.model_proto = self._get_model_proto()
    self.model_id = self.model_proto.id
    self.model_version_id = None
    self.inference_compute_info = self._get_inference_compute_info()
    self.is_v3 = True  # Do model build for v3

  def create_model_instance(self, load_model=True):
    """
    Create an instance of the model class, as specified in the config file.
    """
    # look for default model.py file location
    for loc in ["model.py", "1/model.py"]:
      model_file = os.path.join(self.folder, loc)
      if os.path.exists(model_file):
        break
    if not os.path.exists(model_file):
      raise Exception("Model file not found.")

    module_name = os.path.basename(model_file).replace(".py", "")

    spec = importlib.util.spec_from_file_location(module_name, model_file)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    # Find all classes in the model.py file that are subclasses of ModelClass
    classes = [
        cls for _, cls in inspect.getmembers(module, inspect.isclass)
        if issubclass(cls, ModelClass) and cls.__module__ == module.__name__
    ]
    #  Ensure there is exactly one subclass of BaseRunner in the model.py file
    if len(classes) != 1:
      # check for old inheritence structure, ModelRunner used to be a ModelClass
      runner_classes = [
          cls for _, cls in inspect.getmembers(module, inspect.isclass)
          if cls.__module__ == module.__name__ and any(c.__name__ == 'ModelRunner'
                                                       for c in cls.__bases__)
      ]
      if runner_classes and len(runner_classes) == 1:
        raise Exception(
            f'Could not determine model class.'
            f' Models should now inherit from {ModelClass.__module__}.ModelClass, not ModelRunner.'
            f' Please update your model "{runner_classes[0].__name__}" to inherit from ModelClass.'
        )
      raise Exception(
          "Could not determine model class. There should be exactly one model inheriting from ModelClass defined in the model.py"
      )
    model_class = classes[0]

    # initialize the model
    model = model_class()
    if load_model:
      model.load_model()
    return model

  def _validate_folder(self, folder):
    if folder == ".":
      folder = ""  # will getcwd() next which ends with /
    if not os.path.isabs(folder):
      folder = os.path.join(os.getcwd(), folder)
    logger.info(f"Validating folder: {folder}")
    if not os.path.exists(folder):
      raise FileNotFoundError(f"Folder {folder} not found, please provide a valid folder path")
    files = os.listdir(folder)
    assert "config.yaml" in files, "config.yaml not found in the folder"
    # If just downloading we don't need requirements.txt or the python code, we do need the
    # 1/ folder to put 1/checkpoints into.
    assert "1" in files, "Subfolder '1' not found in the folder"
    if not self.download_validation_only:
      assert "requirements.txt" in files, "requirements.txt not found in the folder"
      subfolder_files = os.listdir(os.path.join(folder, '1'))
      assert 'model.py' in subfolder_files, "model.py not found in the folder"
    return folder

  @staticmethod
  def _load_config(config_file: str):
    with open(config_file, 'r') as file:
      config = yaml.safe_load(file)
    return config

  def _validate_config_checkpoints(self):
    """
    Validates the checkpoints section in the config file.
    :return: loader_type the type of loader or None if no checkpoints.
    :return: repo_id location of checkpoint.
    :return: hf_token token to access checkpoint.
    """
    assert "type" in self.config.get("checkpoints"), "No loader type specified in the config file"
    loader_type = self.config.get("checkpoints").get("type")
    if not loader_type:
      logger.info("No loader type specified in the config file for checkpoints")
      return None, None, None
    assert loader_type == "huggingface", "Only huggingface loader supported for now"
    if loader_type == "huggingface":
      assert "repo_id" in self.config.get("checkpoints"), "No repo_id specified in the config file"
      repo_id = self.config.get("checkpoints").get("repo_id")

      # get from config.yaml otherwise fall back to HF_TOKEN env var.
      hf_token = self.config.get("checkpoints").get("hf_token", os.environ.get("HF_TOKEN", None))
      return loader_type, repo_id, hf_token

  def _check_app_exists(self):
    resp = self.client.STUB.GetApp(service_pb2.GetAppRequest(user_app_id=self.client.user_app_id))
    if resp.status.code == status_code_pb2.SUCCESS:
      return True
    logger.error(
        f"Error checking API {self._base_api} for user app {self.client.user_app_id.user_id}/{self.client.user_app_id.app_id}. Error code: {resp.status.code}"
    )
    return False

  def _validate_config_model(self):
    assert "model" in self.config, "model section not found in the config file"
    model = self.config.get('model')
    assert "user_id" in model, "user_id not found in the config file"
    assert "app_id" in model, "app_id not found in the config file"
    assert "model_type_id" in model, "model_type_id not found in the config file"
    assert "id" in model, "model_id not found in the config file"
    if '.' in model.get('id'):
      logger.error(
          "Model ID cannot contain '.', please remove it from the model_id in the config file")
      sys.exit(1)

    assert model.get('user_id') != "", "user_id cannot be empty in the config file"
    assert model.get('app_id') != "", "app_id cannot be empty in the config file"
    assert model.get('model_type_id') != "", "model_type_id cannot be empty in the config file"
    assert model.get('id') != "", "model_id cannot be empty in the config file"

    if not self._check_app_exists():
      logger.error(
          f"App {self.client.user_app_id.app_id} not found for user {self.client.user_app_id.user_id}"
      )
      sys.exit(1)

  def _validate_config(self):
    if not self.download_validation_only:
      self._validate_config_model()

      assert "inference_compute_info" in self.config, "inference_compute_info not found in the config file"

      if self.config.get("concepts"):
        model_type_id = self.config.get('model').get('model_type_id')
        assert model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE, f"Model type {model_type_id} not supported for concepts"

    if self.config.get("checkpoints"):
      loader_type, _, hf_token = self._validate_config_checkpoints()

      if loader_type == "huggingface" and hf_token:
        is_valid_token = HuggingFaceLoader.validate_hftoken(hf_token)
        if not is_valid_token:
          logger.error(
              "Invalid Hugging Face token provided in the config file, this might cause issues with downloading the restricted model checkpoints."
          )
          logger.info("Continuing without Hugging Face token")

  @staticmethod
  def _get_tar_file_content_size(tar_file_path):
    """
    Calculates the total size of the contents of a tar file.

    Args:
      tar_file_path (str): The path to the tar file.

    Returns:
      int: The total size of the contents in bytes.
    """
    total_size = 0
    with tarfile.open(tar_file_path, 'r') as tar:
      for member in tar:
        if member.isfile():
          total_size += member.size
    return total_size

  @property
  def client(self):
    if self._client is None:
      assert "model" in self.config, "model info not found in the config file"
      model = self.config.get('model')
      assert "user_id" in model, "user_id not found in the config file"
      assert "app_id" in model, "app_id not found in the config file"
      # The owner of the model and the app.
      user_id = model.get('user_id')
      app_id = model.get('app_id')

      self._base_api = os.environ.get('CLARIFAI_API_BASE', 'https://api.clarifai.com')
      self._client = BaseClient(user_id=user_id, app_id=app_id, base=self._base_api)

    return self._client

  @property
  def model_url(self):
    url_helper = ClarifaiUrlHelper(self._client.auth_helper)
    if self.model_version_id is not None:
      return url_helper.clarifai_url(self.client.user_app_id.user_id,
                                     self.client.user_app_id.app_id, "models", self.model_id)
    else:
      return url_helper.clarifai_url(self.client.user_app_id.user_id,
                                     self.client.user_app_id.app_id, "models", self.model_id,
                                     self.model_version_id)

  def _get_model_proto(self):
    assert "model" in self.config, "model info not found in the config file"
    model = self.config.get('model')

    assert "model_type_id" in model, "model_type_id not found in the config file"
    assert "id" in model, "model_id not found in the config file"
    assert "user_id" in model, "user_id not found in the config file"
    assert "app_id" in model, "app_id not found in the config file"

    model_proto = json_format.ParseDict(model, resources_pb2.Model())

    return model_proto

  def _get_inference_compute_info(self):
    assert ("inference_compute_info" in self.config
           ), "inference_compute_info not found in the config file"
    inference_compute_info = self.config.get('inference_compute_info')
    return json_format.ParseDict(inference_compute_info, resources_pb2.ComputeInfo())

  def check_model_exists(self):
    resp = self.client.STUB.GetModel(
        service_pb2.GetModelRequest(
            user_app_id=self.client.user_app_id, model_id=self.model_proto.id))
    if resp.status.code == status_code_pb2.SUCCESS:
      return True
    return False

  def maybe_create_model(self):
    if self.check_model_exists():
      logger.info(
          f"Model '{self.client.user_app_id.user_id}/{self.client.user_app_id.app_id}/models/{self.model_proto.id}' already exists, "
          f"will create a new version for it.")
      return

    request = service_pb2.PostModelsRequest(
        user_app_id=self.client.user_app_id,
        models=[self.model_proto],
    )
    return self.client.STUB.PostModels(request)

  def _match_req_line(self, line):
    line = line.strip()
    if not line or line.startswith('#'):
      return None, None
    # split on whitespace followed by #
    line = re.split(r'\s+#', line)[0]
    if "==" in line:
      pkg, version = line.split("==")
    elif ">=" in line:
      pkg, version = line.split(">=")
    elif ">" in line:
      pkg, version = line.split(">")
    elif "<=" in line:
      pkg, version = line.split("<=")
    elif "<" in line:
      pkg, version = line.split("<")
    else:
      pkg, version = line, None  # No version specified
    for dep in dependencies:
      if dep == pkg:
        if dep == 'torch' and line.find(
            'whl/cpu') > 0:  # Ignore torch-cpu whl files, use base mage.
          return None, None
        return dep.strip(), version.strip() if version else None
    return None, None

  def _parse_requirements(self):
    dependencies_version = {}
    with open(os.path.join(self.folder, 'requirements.txt'), 'r') as file:
      for line in file:
        # Skip empty lines and comments
        dependency, version = self._match_req_line(line)
        if dependency is None:
          continue
        dependencies_version[dependency] = version if version else None
    return dependencies_version

  def create_dockerfile(self):
    dockerfile_template = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        'dockerfile_template',
        'Dockerfile.template',
    )

    with open(dockerfile_template, 'r') as template_file:
      dockerfile_template = template_file.read()

    dockerfile_template = Template(dockerfile_template)

    # Get the Python version from the config file
    build_info = self.config.get('build_info', {})
    if 'python_version' in build_info:
      python_version = build_info['python_version']
      if python_version not in AVAILABLE_PYTHON_IMAGES:
        logger.error(
            f"Python version {python_version} not supported, please use one of the following versions: {AVAILABLE_PYTHON_IMAGES}"
        )
        return
      logger.info(
          f"Using Python version {python_version} from the config file to build the Dockerfile")
    else:
      logger.info(
          f"Python version not found in the config file, using default Python version: {DEFAULT_PYTHON_VERSION}"
      )
      python_version = DEFAULT_PYTHON_VERSION

    # This is always the final image used for runtime.
    final_image = PYTHON_BASE_IMAGE.format(python_version=python_version)
    downloader_image = PYTHON_BASE_IMAGE.format(python_version=python_version)

    # Parse the requirements.txt file to determine the base image
    dependencies = self._parse_requirements()
    if 'torch' in dependencies and dependencies['torch']:
      torch_version = dependencies['torch']

      # Sort in reverse so that newer cuda versions come first and are preferred.
      for image in sorted(AVAILABLE_TORCH_IMAGES, reverse=True):
        if torch_version in image and f'py{python_version}' in image:
          cuda_version = image.split('-')[-1].replace('cuda', '')
          final_image = TORCH_BASE_IMAGE.format(
              torch_version=torch_version,
              python_version=python_version,
              cuda_version=cuda_version,
          )
          logger.info(f"Using Torch version {torch_version} base image to build the Docker image")
          break

    if 'clarifai' not in dependencies:
      raise Exception(
          f"clarifai not found in requirements.txt, please add clarifai to the requirements.txt file with a fixed version. Current version is clarifai=={CLIENT_VERSION}"
      )
    clarifai_version = dependencies['clarifai']
    if not clarifai_version:
      logger.warn(
          f"clarifai version not found in requirements.txt, using the latest version {CLIENT_VERSION}"
      )
      clarifai_version = CLIENT_VERSION
      lines = []
      with open(os.path.join(self.folder, 'requirements.txt'), 'r') as file:
        for line in file:
          # if the line without whitespace is "clarifai"
          dependency, version = self._match_req_line(line)
          if dependency and dependency == "clarifai":
            lines.append(line.replace("clarifai", f"clarifai=={CLIENT_VERSION}"))
          else:
            lines.append(line)
      with open(os.path.join(self.folder, 'requirements.txt'), 'w') as file:
        file.writelines(lines)
      logger.warn(f"Updated requirements.txt to have clarifai=={CLIENT_VERSION}")

    # Replace placeholders with actual values
    dockerfile_content = dockerfile_template.safe_substitute(
        name='main',
        FINAL_IMAGE=final_image,  # for pip requirements
        DOWNLOADER_IMAGE=downloader_image,  # for downloading checkpoints
        CLARIFAI_VERSION=clarifai_version,  # for clarifai
    )

    # Write Dockerfile
    with open(os.path.join(self.folder, 'Dockerfile'), 'w') as dockerfile:
      dockerfile.write(dockerfile_content)

  @property
  def checkpoint_path(self):
    return self._checkpoint_path(self.folder)

  def _checkpoint_path(self, folder):
    return os.path.join(folder, self.checkpoint_suffix)

  @property
  def checkpoint_suffix(self):
    return '1/checkpoints'

  @property
  def tar_file(self):
    return f"{self.folder}.tar.gz"

  def download_checkpoints(self, checkpoint_path_override: str = None):
    """
    Downloads the checkpoints specified in the config file.

    :param checkpoint_path_override: The path to download the checkpoints to. If not provided, the
    default path is used based on the folder ModelUploader was initialized with. The
    checkpoint_suffix will be appended to the path.
    """
    if not self.config.get("checkpoints"):
      logger.info("No checkpoints specified in the config file")
      return True

    loader_type, repo_id, hf_token = self._validate_config_checkpoints()

    success = True
    if loader_type == "huggingface":
      loader = HuggingFaceLoader(repo_id=repo_id, token=hf_token)
      path = self._checkpoint_path(
          checkpoint_path_override) if checkpoint_path_override else self.checkpoint_path
      success = loader.download_checkpoints(path)

    if loader_type:
      if not success:
        logger.error(f"Failed to download checkpoints for model {repo_id}")
        sys.exit(1)
      else:
        logger.info(f"Downloaded checkpoints for model {repo_id}")
    return success

  def _concepts_protos_from_concepts(self, concepts):
    concept_protos = []
    for concept in concepts:
      concept_protos.append(resources_pb2.Concept(
          id=str(concept[0]),
          name=concept[1],
      ))
    return concept_protos

  def hf_labels_to_config(self, labels, config_file):
    with open(config_file, 'r') as file:
      config = yaml.safe_load(file)
    model = config.get('model')
    model_type_id = model.get('model_type_id')
    assert model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE, f"Model type {model_type_id} not supported for concepts"
    concept_protos = self._concepts_protos_from_concepts(labels)

    config['concepts'] = [{'id': concept.id, 'name': concept.name} for concept in concept_protos]

    with open(config_file, 'w') as file:
      yaml.dump(config, file, sort_keys=False)
    concepts = config.get('concepts')
    logger.info(f"Updated config.yaml with {len(concepts)} concepts.")

  def get_model_version_proto(self):

    model_version_proto = resources_pb2.ModelVersion(
        pretrained_model_config=resources_pb2.PretrainedModelConfig(),
        inference_compute_info=self.inference_compute_info,
    )

    model_type_id = self.config.get('model').get('model_type_id')
    if model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE:

      if 'concepts' in self.config:
        labels = self.config.get('concepts')
        logger.info(f"Found {len(labels)} concepts in the config file.")
        for concept in labels:
          concept_proto = json_format.ParseDict(concept, resources_pb2.Concept())
          model_version_proto.output_info.data.concepts.append(concept_proto)
      elif self.config.get("checkpoints") and HuggingFaceLoader.validate_concept(
          self.checkpoint_path):
        labels = HuggingFaceLoader.fetch_labels(self.checkpoint_path)
        logger.info(f"Found {len(labels)} concepts from the model checkpoints.")
        # sort the concepts by id and then update the config file
        labels = sorted(labels.items(), key=lambda x: int(x[0]))

        config_file = os.path.join(self.folder, 'config.yaml')
        try:
          self.hf_labels_to_config(labels, config_file)
        except Exception as e:
          logger.error(f"Failed to update the config.yaml file with the concepts: {e}")

        model_version_proto.output_info.data.concepts.extend(
            self._concepts_protos_from_concepts(labels))
    return model_version_proto

  def upload_model_version(self, download_checkpoints):
    file_path = f"{self.folder}.tar.gz"
    logger.debug(f"Will tar it into file: {file_path}")

    model_type_id = self.config.get('model').get('model_type_id')

    if (model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE) and 'concepts' not in self.config:
      logger.info(
          f"Model type {model_type_id} requires concepts to be specified in the config.yaml file.."
      )
      if self.config.get("checkpoints"):
        logger.info(
            "Checkpoints specified in the config.yaml file, will download the HF model's config.json file to infer the concepts."
        )

        if not download_checkpoints and not HuggingFaceLoader.validate_config(
            self.checkpoint_path):

          input(
              "Press Enter to download the HuggingFace model's config.json file to infer the concepts and continue..."
          )
          loader_type, repo_id, hf_token = self._validate_config_checkpoints()
          if loader_type == "huggingface":
            loader = HuggingFaceLoader(repo_id=repo_id, token=hf_token)
            loader.download_config(self.checkpoint_path)

      else:
        logger.error(
            "No checkpoints specified in the config.yaml file to infer the concepts. Please either specify the concepts directly in the config.yaml file or include a checkpoints section to download the HF model's config.json file to infer the concepts."
        )
        return

    model_version_proto = self.get_model_version_proto()

    def filter_func(tarinfo):
      name = tarinfo.name
      exclude = [self.tar_file, "*~"]
      if not download_checkpoints:
        exclude.append(self.checkpoint_suffix)
      return None if any(name.endswith(ex) for ex in exclude) else tarinfo

    with tarfile.open(self.tar_file, "w:gz") as tar:
      tar.add(self.folder, arcname=".", filter=filter_func)
    logger.debug("Tarring complete, about to start upload.")

    file_size = os.path.getsize(self.tar_file)
    logger.debug(f"Size of the tar is: {file_size} bytes")

    self.storage_request_size = self._get_tar_file_content_size(file_path)
    if not download_checkpoints and self.config.get("checkpoints"):
      # Get the checkpoint size to add to the storage request.
      # First check for the env variable, then try querying huggingface. If all else fails, use the default.
      checkpoint_size = os.environ.get('CHECKPOINT_SIZE_BYTES', 0)
      if not checkpoint_size:
        _, repo_id, _ = self._validate_config_checkpoints()
        checkpoint_size = HuggingFaceLoader.get_huggingface_checkpoint_total_size(repo_id)
      if not checkpoint_size:
        checkpoint_size = self.DEFAULT_CHECKPOINT_SIZE
      self.storage_request_size += checkpoint_size

    self.maybe_create_model()
    if not self.check_model_exists():
      logger.error(f"Failed to create model: {self.model_proto.id}")
      sys.exit(1)

    for response in self.client.STUB.PostModelVersionsUpload(
        self.model_version_stream_upload_iterator(model_version_proto, file_path),):
      percent_completed = 0
      if response.status.code == status_code_pb2.UPLOAD_IN_PROGRESS:
        percent_completed = response.status.percent_completed
      details = response.status.details

      _clear_line()
      print(
          f"Status: {response.status.description}, "
          f"Progress: {percent_completed}% - {details} ",
          f"request_id: {response.status.req_id}",
          end='\r',
          flush=True)
    if response.status.code != status_code_pb2.MODEL_BUILDING:
      logger.error(f"Failed to upload model version: {response}")
      return
    self.model_version_id = response.model_version_id
    logger.info(f"Created Model Version ID: {self.model_version_id}")
    logger.info(f"Full url to that version is: {self.model_url}")
    try:
      self.monitor_model_build()
    finally:
      if os.path.exists(self.tar_file):
        logger.debug(f"Cleaning up upload file: {self.tar_file}")
        os.remove(self.tar_file)

  def model_version_stream_upload_iterator(self, model_version_proto, file_path):
    yield self.init_upload_model_version(model_version_proto, file_path)
    with open(file_path, "rb") as f:
      file_size = os.path.getsize(file_path)
      chunk_size = int(127 * 1024 * 1024)  # 127MB chunk size
      num_chunks = (file_size // chunk_size) + 1
      logger.info("Uploading file...")
      logger.debug(f"File size: {file_size}")
      logger.debug(f"Chunk size: {chunk_size}")
      logger.debug(f"Number of chunks: {num_chunks}")
      read_so_far = 0
      for part_id in range(num_chunks):
        try:
          chunk_size = min(chunk_size, file_size - read_so_far)
          chunk = f.read(chunk_size)
          if not chunk:
            break
          read_so_far += len(chunk)
          yield service_pb2.PostModelVersionsUploadRequest(
              content_part=resources_pb2.UploadContentPart(
                  data=chunk,
                  part_number=part_id + 1,
                  range_start=read_so_far,
              ))
        except Exception as e:
          logger.exception(f"\nError uploading file: {e}")
          break

    if read_so_far == file_size:
      logger.info("Upload complete!")

  def init_upload_model_version(self, model_version_proto, file_path):
    file_size = os.path.getsize(file_path)
    logger.debug(f"Uploading model version of model {self.model_proto.id}")
    logger.debug(f"Using file '{os.path.basename(file_path)}' of size: {file_size} bytes")
    result = service_pb2.PostModelVersionsUploadRequest(
        upload_config=service_pb2.PostModelVersionsUploadConfig(
            user_app_id=self.client.user_app_id,
            model_id=self.model_proto.id,
            model_version=model_version_proto,
            total_size=file_size,
            storage_request_size=self.storage_request_size,
            is_v3=self.is_v3,
        ))
    return result

  def get_model_build_logs(self):
    logs_request = service_pb2.ListLogEntriesRequest(
        log_type="builder",
        user_app_id=self.client.user_app_id,
        model_id=self.model_proto.id,
        model_version_id=self.model_version_id,
        page=1,
        per_page=50)
    response = self.client.STUB.ListLogEntries(logs_request)

    return response

  def monitor_model_build(self):
    st = time.time()
    seen_logs = set()  # To avoid duplicate log messages
    while True:
      resp = self.client.STUB.GetModelVersion(
          service_pb2.GetModelVersionRequest(
              user_app_id=self.client.user_app_id,
              model_id=self.model_proto.id,
              version_id=self.model_version_id,
          ))
      status_code = resp.model_version.status.code
      logs = self.get_model_build_logs()
      for log_entry in logs.log_entries:
        if log_entry.url not in seen_logs:
          seen_logs.add(log_entry.url)
          logger.info(f"{escape(log_entry.message.strip())}")
      if status_code == status_code_pb2.MODEL_BUILDING:
        print(f"Model is building... (elapsed {time.time() - st:.1f}s)", end='\r', flush=True)

        # Fetch and display the logs
        time.sleep(1)
      elif status_code == status_code_pb2.MODEL_TRAINED:
        logger.info("Model build complete!")
        logger.info(f"Build time elapsed {time.time() - st:.1f}s)")
        logger.info(f"Check out the model at {self.model_url} version: {self.model_version_id}")
        return True
      else:
        logger.info(
            f"\nModel build failed with status: {resp.model_version.status} and response {resp}")
        return False


def upload_model(folder, download_checkpoints, skip_dockerfile):
  builder = ModelBuilder(folder)
  if download_checkpoints:
    builder.download_checkpoints()
  if not skip_dockerfile:
    builder.create_dockerfile()
  exists = builder.check_model_exists()
  if exists:
    logger.info(
        f"Model already exists at {builder.model_url}, this upload will create a new version for it."
    )
  else:
    logger.info(f"New model will be created at {builder.model_url} with it's first version.")

  input("Press Enter to continue...")
  builder.upload_model_version(download_checkpoints)



import click
from clarifai.cli.base import cli
from clarifai.client.compute_cluster import ComputeCluster
from clarifai.utils.cli import display_co_resources, dump_yaml, from_yaml


@cli.group(['nodepool', 'np'])
def nodepool():
  """Manage Nodepools: create, delete, list"""
  pass


@nodepool.command()
@click.option(
    '-cc_id',
    '--compute_cluster_id',
    required=False,
    help='Compute Cluster ID for the compute cluster to interact with.')
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=True,
    help='Path to the nodepool config file.')
@click.option(
    '-np_id', '--nodepool_id', required=False, help='New Nodepool ID for the nodepool to create.')
@click.pass_context
def create(ctx, compute_cluster_id, config, nodepool_id):
  """Create a new Nodepool with the given config file."""

  nodepool_config = from_yaml(config)
  if not compute_cluster_id:
    if 'compute_cluster' not in nodepool_config['nodepool']:
      click.echo(
          "Please provide a compute cluster ID either in the config file or using --compute_cluster_id flag",
          err=True)
      return
    compute_cluster_id = nodepool_config['nodepool']['compute_cluster']['id']
  else:
    if 'compute_cluster' not in nodepool_config['nodepool']:
      nodepool_config['nodepool']['compute_cluster']['id'] = compute_cluster_id
      dump_yaml(config, nodepool_config)

  compute_cluster = ComputeCluster(
      compute_cluster_id=compute_cluster_id,
      user_id=ctx.obj['user_id'],
      pat=ctx.obj['pat'],
      base_url=ctx.obj['base_url'])
  if nodepool_id:
    compute_cluster.create_nodepool(config, nodepool_id=nodepool_id)
  else:
    compute_cluster.create_nodepool(config)


@nodepool.command()
@click.option(
    '-cc_id',
    '--compute_cluster_id',
    required=True,
    help='Compute Cluster ID for the compute cluster to interact with.')
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=16)
@click.pass_context
def list(ctx, compute_cluster_id, page_no, per_page):
  """List all nodepools for the user."""

  compute_cluster = ComputeCluster(
      compute_cluster_id=compute_cluster_id,
      user_id=ctx.obj['user_id'],
      pat=ctx.obj['pat'],
      base_url=ctx.obj['base_url'])
  response = compute_cluster.list_nodepools(page_no, per_page)
  display_co_resources(response, "Nodepool")


@nodepool.command()
@click.option(
    '-cc_id',
    '--compute_cluster_id',
    required=True,
    help='Compute Cluster ID for the compute cluster to interact with.')
@click.option('-np_id', '--nodepool_id', help='Nodepool ID of the user to delete.')
@click.pass_context
def delete(ctx, compute_cluster_id, nodepool_id):
  """Deletes a nodepool for the user."""

  compute_cluster = ComputeCluster(
      compute_cluster_id=compute_cluster_id,
      user_id=ctx.obj['user_id'],
      pat=ctx.obj['pat'],
      base_url=ctx.obj['base_url'])
  compute_cluster.delete_nodepools([nodepool_id])



import click

from clarifai.cli.base import cli


@cli.group(['model'])
def model():
  """Manage models: upload, test locally, run locally, predict, and more"""


@model.command()
@click.option(
    '--model_path',
    type=click.Path(exists=True),
    required=True,
    help='Path to the model directory.')
@click.option(
    '--download_checkpoints',
    is_flag=True,
    help=
    'Flag to download checkpoints before uploading and including them in the tar file that is uploaded. Defaults to False, which will attempt to download them at docker build time.',
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help=
    'Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile.',
)
def upload(model_path, download_checkpoints, skip_dockerfile):
  """Upload a model to Clarifai."""
  from clarifai.runners.models.model_builder import upload_model
  upload_model(model_path, download_checkpoints, skip_dockerfile)


@model.command()
@click.option(
    '--model_path',
    type=click.Path(exists=True),
    required=True,
    help='Path to the model directory.')
@click.option(
    '--out_path',
    type=click.Path(exists=False),
    required=False,
    default=None,
    help=
    'Option path to write the checkpoints to. This will place them in {out_path}/ If not provided it will default to {model_path}/1/checkpoints where the config.yaml is read..'
)
def download_checkpoints(model_path, out_path):
  """Download checkpoints from external source to local model_path"""

  from clarifai.runners.models.model_builder import ModelBuilder
  builder = ModelBuilder(model_path, download_validation_only=True)
  builder.download_checkpoints(out_path)


@model.command()
@click.option(
    '--model_path',
    type=click.Path(exists=True),
    required=True,
    help='Path to the model directory.')
@click.option(
    '--mode',
    type=click.Choice(['env', 'container'], case_sensitive=False),
    default='env',
    show_default=True,
    help=
    'Specify how to test the model locally: "env" for virtual environment or "container" for Docker container. Defaults to "env".'
)
@click.option(
    '--keep_env',
    is_flag=True,
    help=
    'Keep the virtual environment after testing the model locally (applicable for virtualenv mode). Defaults to False.'
)
@click.option(
    '--keep_image',
    is_flag=True,
    help=
    'Keep the Docker image after testing the model locally (applicable for container mode). Defaults to False.'
)
def test_locally(model_path, keep_env=False, keep_image=False, mode='env'):
  """Test model locally."""
  try:
    from clarifai.runners.models import model_run_locally
    if mode == 'env' and keep_image:
      raise ValueError("'keep_image' is applicable only for 'container' mode")
    if mode == 'container' and keep_env:
      raise ValueError("'keep_env' is applicable only for 'env' mode")

    if mode == "env":
      click.echo("Testing model locally in a virtual environment...")
      model_run_locally.main(model_path, run_model_server=False, keep_env=keep_env)
    elif mode == "container":
      click.echo("Testing model locally inside a container...")
      model_run_locally.main(
          model_path, inside_container=True, run_model_server=False, keep_image=keep_image)
    click.echo("Model tested successfully.")
  except Exception as e:
    click.echo(f"Failed to test model locally: {e}", err=True)


@model.command()
@click.option(
    '--model_path',
    type=click.Path(exists=True),
    required=True,
    help='Path to the model directory.')
@click.option(
    '--port',
    '-p',
    type=int,
    default=8000,
    show_default=True,
    help="The port to host the gRPC server for running the model locally. Defaults to 8000.")
@click.option(
    '--mode',
    type=click.Choice(['env', 'container'], case_sensitive=False),
    default='env',
    show_default=True,
    help=
    'Specifies how to run the model: "env" for virtual environment or "container" for Docker container. Defaults to "env".'
)
@click.option(
    '--keep_env',
    is_flag=True,
    help=
    'Keep the virtual environment after testing the model locally (applicable for virtualenv mode). Defaults to False.'
)
@click.option(
    '--keep_image',
    is_flag=True,
    help=
    'Keep the Docker image after testing the model locally (applicable for container mode). Defaults to False.'
)
def run_locally(model_path, port, mode, keep_env, keep_image):
  """Run the model locally and start a gRPC server to serve the model."""
  try:
    from clarifai.runners.models import model_run_locally
    if mode == 'env' and keep_image:
      raise ValueError("'keep_image' is applicable only for 'container' mode")
    if mode == 'container' and keep_env:
      raise ValueError("'keep_env' is applicable only for 'env' mode")

    if mode == "env":
      click.echo("Running model locally in a virtual environment...")
      model_run_locally.main(model_path, run_model_server=True, keep_env=keep_env, port=port)
    elif mode == "container":
      click.echo("Running model locally inside a container...")
      model_run_locally.main(
          model_path,
          inside_container=True,
          run_model_server=True,
          port=port,
          keep_image=keep_image)
    click.echo(f"Model server started locally from {model_path} in {mode} mode.")
  except Exception as e:
    click.echo(f"Failed to starts model server locally: {e}", err=True)


@model.command()
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=False,
    help='Path to the model predict config file.')
@click.option('--model_id', required=False, help='Model ID of the model used to predict.')
@click.option('--user_id', required=False, help='User ID of the model used to predict.')
@click.option('--app_id', required=False, help='App ID of the model used to predict.')
@click.option('--model_url', required=False, help='Model URL of the model used to predict.')
@click.option('--file_path', required=False, help='File path of file for the model to predict')
@click.option('--url', required=False, help='URL to the file for the model to predict')
@click.option('--bytes', required=False, help='Bytes to the file for the model to predict')
@click.option('--input_type', required=False, help='Type of input')
@click.option(
    '-cc_id',
    '--compute_cluster_id',
    required=False,
    help='Compute Cluster ID to use for the model')
@click.option('-np_id', '--nodepool_id', required=False, help='Nodepool ID to use for the model')
@click.option(
    '-dpl_id', '--deployment_id', required=False, help='Deployment ID to use for the model')
@click.option(
    '--inference_params', required=False, default='{}', help='Inference parameters to override')
@click.option('--output_config', required=False, default='{}', help='Output config to override')
@click.pass_context
def predict(ctx, config, model_id, user_id, app_id, model_url, file_path, url, bytes, input_type,
            compute_cluster_id, nodepool_id, deployment_id, inference_params, output_config):
  """Predict using the given model"""
  import json

  from clarifai.client.model import Model
  from clarifai.utils.cli import from_yaml
  if config:
    config = from_yaml(config)
    model_id, user_id, app_id, model_url, file_path, url, bytes, input_type, compute_cluster_id, nodepool_id, deployment_id, inference_params, output_config = (
        config.get(k, v)
        for k, v in [('model_id', model_id), ('user_id', user_id), ('app_id', app_id), (
            'model_url', model_url), ('file_path', file_path), ('url', url), ('bytes', bytes), (
                'input_type', input_type), ('compute_cluster_id', compute_cluster_id), (
                    'nodepool_id',
                    nodepool_id), ('deployment_id',
                                   deployment_id), ('inference_params',
                                                    inference_params), ('output_config',
                                                                        output_config)])
  if sum([opt[1] for opt in [(model_id, 1), (user_id, 1), (app_id, 1), (model_url, 3)]
          if opt[0]]) != 3:
    raise ValueError("Either --model_id & --user_id & --app_id or --model_url must be provided.")
  if compute_cluster_id or nodepool_id or deployment_id:
    if sum([
        opt[1] for opt in [(compute_cluster_id, 0.5), (nodepool_id, 0.5), (deployment_id, 1)]
        if opt[0]
    ]) != 1:
      raise ValueError(
          "Either --compute_cluster_id & --nodepool_id or --deployment_id must be provided.")
  if model_url:
    model = Model(url=model_url, pat=ctx.obj['pat'], base_url=ctx.obj['base_url'])
  else:
    model = Model(
        model_id=model_id,
        user_id=user_id,
        app_id=app_id,
        pat=ctx.obj['pat'],
        base_url=ctx.obj['base_url'])

  if inference_params:
    inference_params = json.loads(inference_params)
  if output_config:
    output_config = json.loads(output_config)

  if file_path:
    model_prediction = model.predict_by_filepath(
        filepath=file_path,
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        inference_params=inference_params,
        output_config=output_config)
  elif url:
    model_prediction = model.predict_by_url(
        url=url,
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        inference_params=inference_params,
        output_config=output_config)
  elif bytes:
    bytes = str.encode(bytes)
    model_prediction = model.predict_by_bytes(
        input_bytes=bytes,
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        inference_params=inference_params,
        output_config=output_config)  ## TO DO: Add support for input_id
  click.echo(model_prediction)



import itertools
from typing import Any, Dict, Iterator, List, Tuple

import numpy as np
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.service_pb2 import PostModelOutputsRequest
from google.protobuf import json_format

from ..utils.data_handler import InputDataHandler, OutputDataHandler
from .model_class import ModelClass


class AnyAnyModel(ModelClass):

  def load_model(self):
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    raise NotImplementedError

  def parse_input_request(
      self, input_request: service_pb2.PostModelOutputsRequest) -> Tuple[List[Dict], Dict]:
    list_input_dict = [
        InputDataHandler.from_proto(input).to_python() for input in input_request.inputs
    ]
    inference_params = json_format.MessageToDict(
        input_request.model.model_version.output_info.params)

    return list_input_dict, inference_params

  def convert_output_to_proto(self, outputs: list):
    assert (isinstance(outputs, Iterator) or isinstance(outputs, list) or
            isinstance(outputs, tuple)), "outputs must be an Iterator"
    output_protos = []
    for output in outputs:
      if isinstance(output, OutputDataHandler):
        output = output.proto
      elif isinstance(output, resources_pb2.Output):
        pass
      else:
        raise NotImplementedError
      output_protos.append(output)

    return service_pb2.MultiOutputResponse(outputs=output_protos)

  def predict_wrapper(
      self, request: service_pb2.PostModelOutputsRequest) -> service_pb2.MultiOutputResponse:
    list_dict_input, inference_params = self.parse_input_request(request)
    outputs = self.predict(list_dict_input, inference_parameters=inference_params)
    return self.convert_output_to_proto(outputs)

  def generate_wrapper(
      self, request: PostModelOutputsRequest) -> Iterator[service_pb2.MultiOutputResponse]:
    list_dict_input, inference_params = self.parse_input_request(request)
    outputs = self.generate(list_dict_input, inference_parameters=inference_params)
    for output in outputs:
      yield self.convert_output_to_proto(output)

  def _preprocess_stream(
      self, request: Iterator[PostModelOutputsRequest]) -> Iterator[Tuple[List[Dict], List[Dict]]]:
    """Return generator of processed data (from proto to python) and inference parameters like predict and generate"""
    for i, req in enumerate(request):
      input_data, _ = self.parse_input_request(req)
      yield input_data

  def stream_wrapper(self, request: Iterator[PostModelOutputsRequest]
                    ) -> Iterator[service_pb2.MultiOutputResponse]:
    first_request = next(request)
    _, inference_params = self.parse_input_request(first_request)
    request_iterator = itertools.chain([first_request], request)
    outputs = self.stream(self._preprocess_stream(request_iterator), inference_params)
    for output in outputs:
      yield self.convert_output_to_proto(output)

  def predict(self, input_data: List[Dict],
              inference_parameters: Dict[str, Any] = {}) -> List[OutputDataHandler]:
    """
    Prediction method.

    Args:
    -----
    - input_data: is list of dict where key is input type name.
      * image: np.ndarray
      * text: str
      * audio: bytes

    - inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters.

    Returns:
    --------
      List of OutputDataHandler
    """
    raise NotImplementedError

  def generate(self, input_data: List[Dict],
               inference_parameters: Dict[str, Any] = {}) -> Iterator[List[OutputDataHandler]]:
    """
    Generate method.

    Args:
    -----
    - input_data: is list of dict where key is input type name.
      * image: np.ndarray
      * text: str
      * audio: bytes

    - inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters.

    Yield:
    --------
      List of OutputDataHandler
    """
    raise NotImplementedError

  def stream(self, inputs: Iterator[List[Dict[str, Any]]],
             inference_params: Dict[str, Any]) -> Iterator[List[OutputDataHandler]]:
    """
    Stream method.

    Args:
    -----
    input_request: is an Iterator of Tuple which
    - First element (List[Dict[str, Union[np.ndarray, str, bytes]]]) is list of dict input data type which keys and values are:
        * image: np.ndarray
        * text: str
        * audio: bytes

    - Second element (Dict[str, Union[bool, str, float, int]]): is a dict of inference_parameters

    Yield:
    --------
      List of OutputDataHandler
    """
    raise NotImplementedError


class VisualInputModel(AnyAnyModel):

  def parse_input_request(
      self, input_request: service_pb2.PostModelOutputsRequest) -> Tuple[List[Dict], Dict]:
    list_input_dict = [
        InputDataHandler.from_proto(input).image(format="np") for input in input_request.inputs
    ]
    inference_params = json_format.MessageToDict(
        input_request.model.model_version.output_info.params)

    return list_input_dict, inference_params

  def load_model(self):
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    raise NotImplementedError

  def predict(self, input_data: List[np.ndarray],
              inference_parameters: Dict[str, Any] = {}) -> List[OutputDataHandler]:
    """
    Prediction method.

    Args:
    -----
    - input_data(List[np.ndarray]): is list of image as np.ndarray type
    - inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters.

    Returns:
    --------
      List of OutputDataHandler
    """
    raise NotImplementedError


class TextInputModel(AnyAnyModel):

  def load_model(self):
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    raise NotImplementedError

  def parse_input_request(
      self, input_request: service_pb2.PostModelOutputsRequest) -> Tuple[List[Dict], Dict]:
    list_input_text = [InputDataHandler.from_proto(input).text for input in input_request.inputs]
    inference_params = json_format.MessageToDict(
        input_request.model.model_version.output_info.params)

    return list_input_text, inference_params

  def predict(self, input_data: List[str],
              inference_parameters: Dict[str, Any] = {}) -> List[OutputDataHandler]:
    """
    Prediction method.

    Args:
    -----
    - input_data(List[str]): is list of text as str type
    - inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters.

    Returns:
    --------
      List of OutputDataHandler
    """
    raise NotImplementedError

  def generate(self, input_data: List[str],
               inference_parameters: Dict[str, Any] = {}) -> Iterator[List[OutputDataHandler]]:
    """
    Prediction method.

    Args:
    -----
    - input_data(List[str]): is list of text as str type
    - inference_parameters (Dict[str, Union[bool, str, float, int]]): your inference parameters.

    Yield:
    --------
      List of OutputDataHandler
    """
    raise NotImplementedError

  def stream(self, inputs: Iterator[List[str]],
             inference_params: Dict[str, Any]) -> Iterator[List[OutputDataHandler]]:
    """
    Stream method.

    Args:
    -----
    input_request: is an Iterator of Tuple which
    - First element (List[str]) is list of input text:
    - Second element (Dict[str, Union[bool, str, float, int]]): is a dict of inference_parameters

    Yield:
    --------
      List of OutputDataHandler
    """
    raise NotImplementedError



import logging
import os
import time
import uuid

import pytest
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.user import User
from clarifai.utils.evaluation import EvalResultCompare

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
NOW = uuid.uuid4().hex[:10]
CREATE_APP_ID = f"ci_input_app_{NOW}"
CREATE_DATASET_ID = "ci_input_test_dataset"
CREATE_DATASET_NEW_ID = "ci_input_test_dataset_new"
CREATE_MODEL_ID = "ci_input_test_model_1"
CSV_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.csv"

CLARIFAI_API_BASE = os.environ.get("CLARIFAI_API_BASE", "https://api.clarifai.com")


def client():
  return User(user_id=CREATE_APP_USER_ID, base_url=CLARIFAI_API_BASE)


@pytest.mark.requires_secrets
class TestEval:
  """Tests for model evaluating.
  """

  @classmethod
  def setup_class(self):
    self.app = client().create_app(app_id=CREATE_APP_ID, base_workflow="Universal")
    self.input_object = self.app.inputs()
    self.dataset = self.app.create_dataset(dataset_id=CREATE_DATASET_ID)
    self.dataset_new = self.app.create_dataset(dataset_id=CREATE_DATASET_NEW_ID)
    self.model = self.app.create_model(
        model_id=CREATE_MODEL_ID, model_type_id='embedding-classifier')

  def test_evaluate(self, caplog):
    # Prepare dataset
    self.dataset.upload_from_csv(
        csv_path=CSV_FILE_PATH, input_type='text', csv_type='raw', labels=True)
    dataset_version = self.dataset.create_version()
    self.dataset_new.upload_from_csv(
        csv_path=CSV_FILE_PATH, input_type='text', csv_type='raw', labels=True)
    self.dataset_new.create_version()
    concepts = [concept.id for concept in self.app.list_concepts()]
    # Prepare for training
    self.model.get_params(save_to='tests/assets/model_params_for_eval.yaml')
    assert len(concepts) == 2  #test data upload for training
    self.model.update_params(
        dataset_id=CREATE_DATASET_ID,
        concepts=concepts,
        dataset_version_id=dataset_version.version.id)
    with caplog.at_level(logging.INFO):
      model_version_id = self.model.train()
      assert "Model Training Started" in caplog.text  #test model training

    while True:
      status = self.model.training_status(version_id=model_version_id, training_logs=True)
      if status.code == 21106:  #MODEL_TRAINING_FAILED
        break
      elif status.code == 21100:  #MODEL_TRAINED
        break
      else:
        time.sleep(2)

    # Test eval
    ## Test list eval
    all_evals = self.model.list_evaluations()
    assert len(all_evals) == 0

    ## Test evaluate
    self.model.evaluate(dataset=self.dataset, eval_id="one", eval_info={"use_kfold": False})
    all_evals = self.model.list_evaluations()
    assert len(all_evals) == 1

    ## Test get_eval_by_id
    time.time()
    while True:
      response1 = self.model.get_eval_by_id("one")
      if (response1.status.code == status_code_pb2.MODEL_EVALUATED):
        break
      else:
        time.sleep(2)

    ## Test get_eval_by_dataset
    dataset_eval = self.model.get_eval_by_dataset(dataset=self.dataset)
    assert dataset_eval[0].id == "one" and dataset_eval[0].status.code == status_code_pb2.MODEL_EVALUATED

    ## Test get_raw_eval
    with pytest.raises(Exception) as e:
      self.model.get_raw_eval(dataset=self.dataset)
    assert "method only supports model types ['visual-classifier', 'text-classifier', 'visual-detector']" in str(
        e.value).lower()

    ## Test get_latest_eval
    current_eval = self.model.get_latest_eval(
        label_counts=True,
        test_set=True,
        binary_metrics=True,
        confusion_matrix=True,
        metrics_by_area=True,
        metrics_by_class=True)
    assert current_eval.id == "one" and current_eval.status.code == status_code_pb2.MODEL_EVALUATED

    ## Test EvalResultCompare
    eval_result = EvalResultCompare(
        models=[self.model],
        datasets=[self.dataset, self.dataset_new],
        attempt_evaluate=True,
        eval_info={
            "use_kfold": False
        })
    eval_result.all('tests/eval/', overwrite=True)
    assert os.path.exists('tests/eval/conf-0.5') is True

    # cleanup
    with caplog.at_level(logging.INFO):
      self.model.delete_version(version_id=model_version_id)
      assert "Model Version Deleted" in caplog.text  #test model version deletion
    os.remove('tests/assets/model_params_for_eval.yaml')

  @classmethod
  def teardown_class(self):
    self.app.delete_model(model_id=CREATE_MODEL_ID)
    self.app.delete_dataset(dataset_id=CREATE_DATASET_ID)
    client().delete_app(app_id=CREATE_APP_ID)



import json
import os
import time
from typing import Any, Dict, Generator, Iterator, List, Tuple, Union

import numpy as np
import requests
import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.resources_pb2 import Input, RunnerSelector
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct, Value
from requests.adapters import HTTPAdapter, Retry
from tqdm import tqdm

from clarifai.client.base import BaseClient
from clarifai.client.dataset import Dataset
from clarifai.client.deployment import Deployment
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.client.nodepool import Nodepool
from clarifai.constants.model import (CHUNK_SIZE, MAX_CHUNK_SIZE, MAX_MODEL_PREDICT_INPUTS,
                                      MAX_RANGE_SIZE, MIN_CHUNK_SIZE, MIN_RANGE_SIZE,
                                      MODEL_EXPORT_TIMEOUT, RANGE_SIZE, TRAINABLE_MODEL_TYPES)
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import logger
from clarifai.utils.misc import BackoffIterator
from clarifai.utils.model_train import (find_and_replace_key, params_parser,
                                        response_to_model_params, response_to_param_info,
                                        response_to_templates)

MAX_SIZE_PER_STREAM = int(89_128_960)  # 85GiB
MIN_CHUNK_FOR_UPLOAD_FILE = int(5_242_880)  # 5MiB
MAX_CHUNK_FOR_UPLOAD_FILE = int(5_242_880_000)  # 5GiB


class Model(Lister, BaseClient):
  """Model is a class that provides access to Clarifai API endpoints related to Model information."""

  def __init__(self,
               url: str = None,
               model_id: str = None,
               model_version: Dict = {'id': ""},
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes a Model object.

    Args:
        url (str): The URL to initialize the model object.
        model_id (str): The Model ID to interact with.
        model_version (dict): The Model Version to interact with.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
        **kwargs: Additional keyword arguments to be passed to the Model.
    """
    if url and model_id:
      raise UserError("You can only specify one of url or model_id.")
    if not url and not model_id:
      raise UserError("You must specify one of url or model_id.")
    if url:
      user_id, app_id, _, model_id, model_version_id = ClarifaiUrlHelper.split_clarifai_url(url)
      model_version = {'id': model_version_id}
      kwargs = {'user_id': user_id, 'app_id': app_id}
    self.kwargs = {**kwargs, 'id': model_id, 'model_version': model_version, }
    self.model_info = resources_pb2.Model(**self.kwargs)
    self.logger = logger
    self.training_params = {}
    self.input_types = None
    BaseClient.__init__(
        self,
        user_id=self.user_id,
        app_id=self.app_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  def list_training_templates(self) -> List[str]:
    """Lists all the training templates for the model type.

    Returns:
        templates (List): List of training templates for the model type.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> print(model.list_training_templates())
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    templates = response_to_templates(
        response=response, model_type_id=self.model_info.model_type_id)

    return templates

  def get_params(self, template: str = None, save_to: str = 'params.yaml') -> Dict[str, Any]:
    """Returns the model params for the model type and yaml file.

    Args:
        template (str): The template to use for the model type.
        yaml_file (str): The yaml file to save the model params.

    Returns:
        params (Dict): Dictionary of model params for the model type.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if template is None and self.model_info.model_type_id not in [
        "clusterer", "embedding-classifier"
    ]:
      raise UserError(
          f"Template should be provided for {self.model_info.model_type_id} model type")
    if template is not None and self.model_info.model_type_id in [
        "clusterer", "embedding-classifier"
    ]:
      raise UserError(
          f"Template should not be provided for {self.model_info.model_type_id} model type")

    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    params = response_to_model_params(
        response=response, model_type_id=self.model_info.model_type_id, template=template)
    # yaml file
    assert save_to.endswith('.yaml'), "File extension should be .yaml"
    with open(save_to, 'w') as f:
      yaml.dump(params, f, default_flow_style=False, sort_keys=False)
    # updating the global model params
    self.training_params.update(params)

    return params

  def update_params(self, **kwargs) -> None:
    """Updates the model params for the model.

    Args:
        **kwargs: model params to update.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.update_params(batch_size = 8, dataset_version = 'dataset_version_id')
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if len(self.training_params) == 0:
      raise UserError(
          f"Run 'model.get_params' to get the params for the {self.model_info.model_type_id} model type"
      )
    # getting all the keys in nested dictionary
    all_keys = [key for key in self.training_params.keys()] + [
        key for key in self.training_params.values() if isinstance(key, dict) for key in key
    ]
    # checking if the given params are valid
    if not set(kwargs.keys()).issubset(all_keys):
      raise UserError("Invalid params")
    # updating the global model params
    for key, value in kwargs.items():
      find_and_replace_key(self.training_params, key, value)

  def get_param_info(self, param: str) -> Dict[str, Any]:
    """Returns the param info for the param.

    Args:
        param (str): The param to get the info for.

    Returns:
        param_info (Dict): Dictionary of model param info for the param.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.get_param_info('param')
    """
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if len(self.training_params) == 0:
      raise UserError(
          f"Run 'model.get_params' to get the params for the {self.model_info.model_type_id} model type"
      )

    all_keys = [key for key in self.training_params.keys()] + [
        key for key in self.training_params.values() if isinstance(key, dict) for key in key
    ]
    if param not in all_keys:
      raise UserError(f"Invalid param: '{param}' for model type '{self.model_info.model_type_id}'")
    template = self.training_params['train_params']['template'] if 'template' in all_keys else None

    request = service_pb2.ListModelTypesRequest(user_app_id=self.user_app_id,)
    response = self._grpc_request(self.STUB.ListModelTypes, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    param_info = response_to_param_info(
        response=response,
        model_type_id=self.model_info.model_type_id,
        param=param,
        template=template)

    return param_info

  def train(self, yaml_file: str = None) -> str:
    """Trains the model based on the given yaml file or model params.

    Args:
        yaml_file (str): The yaml file for the model params.

    Returns:
        model_version_id (str): The model version ID for the model.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_params = model.get_params(template='template', yaml_file='model_params.yaml')
        >>> model.train('model_params.yaml')
    """
    if not self.model_info.model_type_id:
      self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")
    if not yaml_file and len(self.training_params) == 0:
      raise UserError("Provide yaml file or run 'model.get_params()'")

    if yaml_file:
      with open(yaml_file, 'r') as file:
        params_dict = yaml.safe_load(file)
    else:
      params_dict = self.training_params
    # getting all the concepts for the model type
    if self.model_info.model_type_id not in ["clusterer", "text-to-text"]:
      concepts = self._list_concepts()
    train_dict = params_parser(params_dict, concepts)
    request = service_pb2.PostModelVersionsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_versions=[resources_pb2.ModelVersion(**train_dict)])
    response = self._grpc_request(self.STUB.PostModelVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Training Started\n%s", response.status)

    return response.model.model_version.id

  def training_status(self, version_id: str = None, training_logs: bool = False) -> Dict[str, str]:
    """Get the training status for the model version. Also stores training logs

    Args:
        version_id (str): The version ID to get the training status for.
        training_logs (bool): Whether to save the training logs in a file.

    Returns:
        training_status (Dict): Dictionary of training status for the model version.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model.training_status(version_id='version_id',training_logs=True)
    """
    if not version_id and not self.model_info.model_version.id:
      raise UserError(
          "Model version ID is missing. Please provide a `model_version` with a valid `id` as an argument or as a URL in the following format: '{user_id}/{app_id}/models/{your_model_id}/model_version_id/{your_version_model_id}' when initializing."
      )

    self.load_info()
    if self.model_info.model_type_id not in TRAINABLE_MODEL_TYPES:
      raise UserError(f"Model type {self.model_info.model_type_id} is not trainable")

    if training_logs:
      try:
        if self.model_info.model_version.train_log:
          log_response = requests.get(self.model_info.model_version.train_log)
          log_response.raise_for_status()  # Check for any HTTP errors
          with open(version_id + '.log', 'wb') as file:
            for chunk in log_response.iter_content(chunk_size=4096):  # 4KB
              file.write(chunk)
          self.logger.info(f"\nTraining logs are saving in '{version_id+'.log'}' file")

      except requests.exceptions.RequestException as e:
        raise Exception(f"An error occurred while getting training logs: {e}")

    return self.model_info.model_version.status

  def delete_version(self, version_id: str) -> None:
    """Deletes a model version for the Model.

    Args:
        version_id (str): The version ID to delete.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model.delete_version(version_id='version_id')
    """
    request = service_pb2.DeleteModelVersionRequest(
        user_app_id=self.user_app_id, model_id=self.id, version_id=version_id)

    response = self._grpc_request(self.STUB.DeleteModelVersion, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Version Deleted\n%s", response.status)

  def create_version(self, **kwargs) -> 'Model':
    """Creates a model version for the Model.

    Args:
        **kwargs: Additional keyword arguments to be passed to Model Version.
          - description (str): The description of the model version.
          - concepts (list[Concept]): The concepts to associate with the model version.
          - output_info (resources_pb2.OutputInfo(): The output info to associate with the model version.

    Returns:
        Model: A Model object for the specified model ID.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url")
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_version = model.create_version(description='model_version_description')
    """
    if self.model_info.model_type_id in TRAINABLE_MODEL_TYPES:
      raise UserError(
          f"{self.model_info.model_type_id} is a trainable model type. Use 'model.train()' to train the model"
      )

    request = service_pb2.PostModelVersionsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_versions=[resources_pb2.ModelVersion(**kwargs)])

    response = self._grpc_request(self.STUB.PostModelVersions, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nModel Version created\n%s", response.status)

    kwargs.update({'app_id': self.app_id, 'user_id': self.user_id})
    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    kwargs = self.process_response_keys(dict_response['model'], 'model')

    return Model(base_url=self.base, pat=self.pat, token=self.token, **kwargs)

  def list_versions(self, page_no: int = None,
                    per_page: int = None) -> Generator['Model', None, None]:
    """Lists all the versions for the model.

    Args:
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Model: Model objects for the versions of the model.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> all_model_versions = list(model.list_versions())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(
        user_app_id=self.user_app_id,
        model_id=self.id,
    )
    all_model_versions_info = self.list_pages_generator(
        self.STUB.ListModelVersions,
        service_pb2.ListModelVersionsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for model_version_info in all_model_versions_info:
      model_version_info['id'] = model_version_info['model_version_id']
      del model_version_info['model_version_id']
      try:
        del model_version_info['train_info']['dataset']['version']['metrics']
      except KeyError:
        pass
      yield Model.from_auth_helper(
          auth=self.auth_helper,
          model_id=self.id,
          **dict(self.kwargs, model_version=model_version_info))

  def predict(self,
              inputs: List[Input],
              runner_selector: RunnerSelector = None,
              inference_params: Dict = {},
              output_config: Dict = {}):
    """Predicts the model based on the given inputs.

    Args:
        inputs (list[Input]): The inputs to predict, must be less than 128.
        runner_selector (RunnerSelector): The runner selector to use for the model.
    """
    if not isinstance(inputs, list):
      raise UserError('Invalid inputs, inputs must be a list of Input objects.')
    if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
      raise UserError(f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}."
                     )  # TODO Use Chunker for inputs len > 128

    self._override_model_version(inference_params, output_config)
    request = service_pb2.PostModelOutputsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        version_id=self.model_version.id,
        inputs=inputs,
        runner_selector=runner_selector,
        model=self.model_info)

    start_time = time.time()
    backoff_iterator = BackoffIterator(10)
    while True:
      response = self._grpc_request(self.STUB.PostModelOutputs, request)

      if response.status.code == status_code_pb2.MODEL_DEPLOYING and \
              time.time() - start_time < 60 * 10:  # 10 minutes
        self.logger.info(f"{self.id} model is still deploying, please wait...")
        time.sleep(next(backoff_iterator))
        continue

      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model Predict failed with response {response.status!r}")
      else:
        break

    return response

  def _check_predict_input_type(self, input_type: str) -> None:
    """Checks if the input type is valid for the model.

    Args:
        input_type (str): The input type to check.
    Returns:
        None
    """
    if not input_type:
      self.load_input_types()
      if len(self.input_types) > 1:
        raise UserError(
            "Model has multiple input types. Please use model.predict() for this multi-modal model."
        )
    else:
      self.input_types = [input_type]
      if self.input_types[0] not in {'image', 'text', 'video', 'audio'}:
        raise UserError(
            f"Got input type {input_type} but expected one of image, text, video, audio.")

  def load_input_types(self) -> None:
    """Loads the input types for the model.

    Returns:
        None

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model.load_input_types()
    """
    if self.input_types:
      return self.input_types
    if self.model_info.model_type_id == "":
      self.load_info()
    request = service_pb2.GetModelTypeRequest(
        user_app_id=self.user_app_id,
        model_type_id=self.model_info.model_type_id,
    )
    response = self._grpc_request(self.STUB.GetModelType, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.input_types = response.model_type.input_fields

  def predict_by_filepath(self,
                          filepath: str,
                          input_type: str = None,
                          compute_cluster_id: str = None,
                          nodepool_id: str = None,
                          deployment_id: str = None,
                          user_id: str = None,
                          inference_params: Dict = {},
                          output_config: Dict = {}):
    """Predicts the model based on the given filepath.

    Args:
        filepath (str): The filepath to predict.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_prediction = model.predict_by_filepath('/path/to/image.jpg')
        >>> model_prediction = model.predict_by_filepath('/path/to/text.txt')
    """
    if not os.path.isfile(filepath):
      raise UserError('Invalid filepath.')

    with open(filepath, "rb") as f:
      file_bytes = f.read()

    return self.predict_by_bytes(file_bytes, input_type, compute_cluster_id, nodepool_id,
                                 deployment_id, user_id, inference_params, output_config)

  def predict_by_bytes(self,
                       input_bytes: bytes,
                       input_type: str = None,
                       compute_cluster_id: str = None,
                       nodepool_id: str = None,
                       deployment_id: str = None,
                       user_id: str = None,
                       inference_params: Dict = {},
                       output_config: Dict = {}):
    """Predicts the model based on the given bytes.

    Args:
        input_bytes (bytes): File Bytes to predict on.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("https://clarifai.com/openai/chat-completion/models/GPT-4")
        >>> model_prediction = model.predict_by_bytes(b'Write a tweet on future of AI',
                                                      inference_params=dict(temperature=str(0.7), max_tokens=30)))
    """
    self._check_predict_input_type(input_type)

    if self.input_types[0] == "image":
      input_proto = Inputs.get_input_from_bytes("", image_bytes=input_bytes)
    elif self.input_types[0] == "text":
      input_proto = Inputs.get_input_from_bytes("", text_bytes=input_bytes)
    elif self.input_types[0] == "video":
      input_proto = Inputs.get_input_from_bytes("", video_bytes=input_bytes)
    elif self.input_types[0] == "audio":
      input_proto = Inputs.get_input_from_bytes("", audio_bytes=input_bytes)

    if deployment_id and (compute_cluster_id or nodepool_id):
      raise UserError(
          "You can only specify one of deployment_id or compute_cluster_id and nodepool_id.")

    runner_selector = None
    if deployment_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with deployment ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Deployment.get_runner_selector(
          user_id=user_id, deployment_id=deployment_id)
    elif compute_cluster_id and nodepool_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with compute cluster ID and nodepool ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Nodepool.get_runner_selector(
          user_id=user_id, compute_cluster_id=compute_cluster_id, nodepool_id=nodepool_id)

    return self.predict(
        inputs=[input_proto],
        runner_selector=runner_selector,
        inference_params=inference_params,
        output_config=output_config)

  def predict_by_url(self,
                     url: str,
                     input_type: str = None,
                     compute_cluster_id: str = None,
                     nodepool_id: str = None,
                     deployment_id: str = None,
                     user_id: str = None,
                     inference_params: Dict = {},
                     output_config: Dict = {}):
    """Predicts the model based on the given URL.

    Args:
        url (str): The URL to predict.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio'.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> model_prediction = model.predict_by_url('url')
    """
    self._check_predict_input_type(input_type)

    if self.input_types[0] == "image":
      input_proto = Inputs.get_input_from_url("", image_url=url)
    elif self.input_types[0] == "text":
      input_proto = Inputs.get_input_from_url("", text_url=url)
    elif self.input_types[0] == "video":
      input_proto = Inputs.get_input_from_url("", video_url=url)
    elif self.input_types[0] == "audio":
      input_proto = Inputs.get_input_from_url("", audio_url=url)

    if deployment_id and (compute_cluster_id or nodepool_id):
      raise UserError(
          "You can only specify one of deployment_id or compute_cluster_id and nodepool_id.")

    runner_selector = None
    if deployment_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with deployment ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Deployment.get_runner_selector(
          user_id=user_id, deployment_id=deployment_id)
    elif compute_cluster_id and nodepool_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with compute cluster ID and nodepool ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Nodepool.get_runner_selector(
          user_id=user_id, compute_cluster_id=compute_cluster_id, nodepool_id=nodepool_id)

    return self.predict(
        inputs=[input_proto],
        runner_selector=runner_selector,
        inference_params=inference_params,
        output_config=output_config)

  def generate(self,
               inputs: List[Input],
               runner_selector: RunnerSelector = None,
               inference_params: Dict = {},
               output_config: Dict = {}):
    """Generate the stream output on model based on the given inputs.

    Args:
        inputs (list[Input]): The inputs to generate, must be less than 128.
        runner_selector (RunnerSelector): The runner selector to use for the model.
        inference_params (dict): The inference params to override.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> stream_response = model.generate(inputs=[input1, input2], runner_selector=runner_selector)
        >>> list_stream_response = [response for response in stream_response]
    """
    if not isinstance(inputs, list):
      raise UserError('Invalid inputs, inputs must be a list of Input objects.')
    if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
      raise UserError(f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}."
                     )  # TODO Use Chunker for inputs len > 128

    self._override_model_version(inference_params, output_config)
    request = service_pb2.PostModelOutputsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        version_id=self.model_version.id,
        inputs=inputs,
        runner_selector=runner_selector,
        model=self.model_info)

    start_time = time.time()
    backoff_iterator = BackoffIterator(10)
    generation_started = False
    while True:
      if generation_started:
        break
      stream_response = self._grpc_request(self.STUB.GenerateModelOutputs, request)
      for response in stream_response:
        if response.status.code == status_code_pb2.MODEL_DEPLOYING and \
                time.time() - start_time < 60 * 10:
          self.logger.info(f"{self.id} model is still deploying, please wait...")
          time.sleep(next(backoff_iterator))
          break
        if response.status.code != status_code_pb2.SUCCESS:
          raise Exception(f"Model Predict failed with response {response.status!r}")
        else:
          if not generation_started:
            generation_started = True
          yield response

  def generate_by_filepath(self,
                           filepath: str,
                           input_type: str = None,
                           compute_cluster_id: str = None,
                           nodepool_id: str = None,
                           deployment_id: str = None,
                           user_id: str = None,
                           inference_params: Dict = {},
                           output_config: Dict = {}):
    """Generate the stream output on model based on the given filepath.

    Args:
        filepath (str): The filepath to predict.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> stream_response = model.generate_by_filepath('/path/to/image.jpg', 'image', deployment_id='deployment_id')
        >>> list_stream_response = [response for response in stream_response]
    """
    if not os.path.isfile(filepath):
      raise UserError('Invalid filepath.')

    with open(filepath, "rb") as f:
      file_bytes = f.read()

    return self.generate_by_bytes(
        input_bytes=file_bytes,
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        user_id=user_id,
        inference_params=inference_params,
        output_config=output_config)

  def generate_by_bytes(self,
                        input_bytes: bytes,
                        input_type: str = None,
                        compute_cluster_id: str = None,
                        nodepool_id: str = None,
                        deployment_id: str = None,
                        user_id: str = None,
                        inference_params: Dict = {},
                        output_config: Dict = {}):
    """Generate the stream output on model based on the given bytes.

    Args:
        input_bytes (bytes): File Bytes to predict on.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("https://clarifai.com/openai/chat-completion/models/GPT-4")
        >>> stream_response = model.generate_by_bytes(b'Write a tweet on future of AI',
                                                      deployment_id='deployment_id',
                                                      inference_params=dict(temperature=str(0.7), max_tokens=30)))
        >>> list_stream_response = [response for response in stream_response]
    """
    self._check_predict_input_type(input_type)

    if self.input_types[0] == "image":
      input_proto = Inputs.get_input_from_bytes("", image_bytes=input_bytes)
    elif self.input_types[0] == "text":
      input_proto = Inputs.get_input_from_bytes("", text_bytes=input_bytes)
    elif self.input_types[0] == "video":
      input_proto = Inputs.get_input_from_bytes("", video_bytes=input_bytes)
    elif self.input_types[0] == "audio":
      input_proto = Inputs.get_input_from_bytes("", audio_bytes=input_bytes)

    if deployment_id and (compute_cluster_id or nodepool_id):
      raise UserError(
          "You can only specify one of deployment_id or compute_cluster_id and nodepool_id.")

    runner_selector = None
    if deployment_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with deployment ID, please provide user_id in the method call."
        )
      runner_selector = Deployment.get_runner_selector(
          user_id=user_id, deployment_id=deployment_id)
    elif compute_cluster_id and nodepool_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with compute cluster ID and nodepool ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Nodepool.get_runner_selector(
          user_id=user_id, compute_cluster_id=compute_cluster_id, nodepool_id=nodepool_id)

    return self.generate(
        inputs=[input_proto],
        runner_selector=runner_selector,
        inference_params=inference_params,
        output_config=output_config)

  def generate_by_url(self,
                      url: str,
                      input_type: str = None,
                      compute_cluster_id: str = None,
                      nodepool_id: str = None,
                      deployment_id: str = None,
                      user_id: str = None,
                      inference_params: Dict = {},
                      output_config: Dict = {}):
    """Generate the stream output on model based on the given URL.

    Args:
        url (str): The URL to predict.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> stream_response = model.generate_by_url('url', deployment_id='deployment_id')
        >>> list_stream_response = [response for response in stream_response]
    """
    self._check_predict_input_type(input_type)

    if self.input_types[0] == "image":
      input_proto = Inputs.get_input_from_url("", image_url=url)
    elif self.input_types[0] == "text":
      input_proto = Inputs.get_input_from_url("", text_url=url)
    elif self.input_types[0] == "video":
      input_proto = Inputs.get_input_from_url("", video_url=url)
    elif self.input_types[0] == "audio":
      input_proto = Inputs.get_input_from_url("", audio_url=url)

    if deployment_id and (compute_cluster_id or nodepool_id):
      raise UserError(
          "You can only specify one of deployment_id or compute_cluster_id and nodepool_id.")

    runner_selector = None
    if deployment_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with deployment ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Deployment.get_runner_selector(
          user_id=user_id, deployment_id=deployment_id)
    elif compute_cluster_id and nodepool_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with compute cluster ID and nodepool ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Nodepool.get_runner_selector(
          user_id=user_id, compute_cluster_id=compute_cluster_id, nodepool_id=nodepool_id)

    return self.generate(
        inputs=[input_proto],
        runner_selector=runner_selector,
        inference_params=inference_params,
        output_config=output_config)

  def _req_iterator(self, input_iterator: Iterator[List[Input]], runner_selector: RunnerSelector):
    for inputs in input_iterator:
      yield service_pb2.PostModelOutputsRequest(
          user_app_id=self.user_app_id,
          model_id=self.id,
          version_id=self.model_version.id,
          inputs=inputs,
          runner_selector=runner_selector,
          model=self.model_info)

  def stream(self,
             inputs: Iterator[List[Input]],
             runner_selector: RunnerSelector = None,
             inference_params: Dict = {},
             output_config: Dict = {}):
    """Generate the stream output on model based on the given stream of inputs.

    Args:
        inputs (Iterator[list[Input]]): stream of inputs to predict, must be less than 128.
        runner_selector (RunnerSelector): The runner selector to use for the model.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url") # Example URL: https://clarifai.com/clarifai/main/models/general-image-recognition
                    or
        >>> model = Model(model_id='model_id', user_id='user_id', app_id='app_id')
        >>> stream_response = model.stream(inputs=inputs, runner_selector=runner_selector)
        >>> list_stream_response = [response for response in stream_response]
    """
    # if not isinstance(inputs, Iterator[List[Input]]):
    #   raise UserError('Invalid inputs, inputs must be a iterator of list of Input objects.')

    self._override_model_version(inference_params, output_config)
    request = self._req_iterator(inputs, runner_selector)

    start_time = time.time()
    backoff_iterator = BackoffIterator(10)
    generation_started = False
    while True:
      if generation_started:
        break
      stream_response = self._grpc_request(self.STUB.StreamModelOutputs, request)
      for response in stream_response:
        if response.status.code == status_code_pb2.MODEL_DEPLOYING and \
                time.time() - start_time < 60 * 10:
          self.logger.info(f"{self.id} model is still deploying, please wait...")
          time.sleep(next(backoff_iterator))
          break
        if response.status.code != status_code_pb2.SUCCESS:
          raise Exception(f"Model Predict failed with response {response.status!r}")
        else:
          if not generation_started:
            generation_started = True
          yield response

  def stream_by_filepath(self,
                         filepath: str,
                         input_type: str = None,
                         compute_cluster_id: str = None,
                         nodepool_id: str = None,
                         deployment_id: str = None,
                         user_id: str = None,
                         inference_params: Dict = {},
                         output_config: Dict = {}):
    """Stream the model output based on the given filepath.

    Args:
        filepath (str): The filepath to predict.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url")
        >>> stream_response = model.stream_by_filepath('/path/to/image.jpg', deployment_id='deployment_id')
        >>> list_stream_response = [response for response in stream_response]
    """
    if not os.path.isfile(filepath):
      raise UserError('Invalid filepath.')

    with open(filepath, "rb") as f:
      file_bytes = f.read()

    return self.stream_by_bytes(
        input_bytes_iterator=iter([file_bytes]),
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        user_id=user_id,
        inference_params=inference_params,
        output_config=output_config)

  def stream_by_bytes(self,
                      input_bytes_iterator: Iterator[bytes],
                      input_type: str = None,
                      compute_cluster_id: str = None,
                      nodepool_id: str = None,
                      deployment_id: str = None,
                      user_id: str = None,
                      inference_params: Dict = {},
                      output_config: Dict = {}):
    """Stream the model output based on the given bytes.

    Args:
        input_bytes_iterator (Iterator[bytes]): Iterator of file bytes to predict on.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("https://clarifai.com/openai/chat-completion/models/GPT-4")
        >>> stream_response = model.stream_by_bytes(iter([b'Write a tweet on future of AI']),
                                                    deployment_id='deployment_id',
                                                    inference_params=dict(temperature=str(0.7), max_tokens=30)))
        >>> list_stream_response = [response for response in stream_response]
    """
    self._check_predict_input_type(input_type)

    def input_generator():
      for input_bytes in input_bytes_iterator:
        if self.input_types[0] == "image":
          yield [Inputs.get_input_from_bytes("", image_bytes=input_bytes)]
        elif self.input_types[0] == "text":
          yield [Inputs.get_input_from_bytes("", text_bytes=input_bytes)]
        elif self.input_types[0] == "video":
          yield [Inputs.get_input_from_bytes("", video_bytes=input_bytes)]
        elif self.input_types[0] == "audio":
          yield [Inputs.get_input_from_bytes("", audio_bytes=input_bytes)]

    if deployment_id and (compute_cluster_id or nodepool_id):
      raise UserError(
          "You can only specify one of deployment_id or compute_cluster_id and nodepool_id.")

    runner_selector = None
    if deployment_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with deployment ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Deployment.get_runner_selector(
          user_id=user_id, deployment_id=deployment_id)
    elif compute_cluster_id and nodepool_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with compute cluster ID and nodepool ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Nodepool.get_runner_selector(
          user_id=user_id, compute_cluster_id=compute_cluster_id, nodepool_id=nodepool_id)

    return self.stream(
        inputs=input_generator(),
        runner_selector=runner_selector,
        inference_params=inference_params,
        output_config=output_config)

  def stream_by_url(self,
                    url_iterator: Iterator[str],
                    input_type: str = None,
                    compute_cluster_id: str = None,
                    nodepool_id: str = None,
                    deployment_id: str = None,
                    user_id: str = None,
                    inference_params: Dict = {},
                    output_config: Dict = {}):
    """Stream the model output based on the given URL.

    Args:
        url_iterator (Iterator[str]): Iterator of URLs to predict.
        input_type (str, optional): The type of input. Can be 'image', 'text', 'video' or 'audio.
        compute_cluster_id (str): The compute cluster ID to use for the model.
        nodepool_id (str): The nodepool ID to use for the model.
        deployment_id (str): The deployment ID to use for the model.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url")
        >>> stream_response = model.stream_by_url(iter(['url']), deployment_id='deployment_id')
        >>> list_stream_response = [response for response in stream_response]
    """
    self._check_predict_input_type(input_type)

    def input_generator():
      for url in url_iterator:
        if self.input_types[0] == "image":
          yield [Inputs.get_input_from_url("", image_url=url)]
        elif self.input_types[0] == "text":
          yield [Inputs.get_input_from_url("", text_url=url)]
        elif self.input_types[0] == "video":
          yield [Inputs.get_input_from_url("", video_url=url)]
        elif self.input_types[0] == "audio":
          yield [Inputs.get_input_from_url("", audio_url=url)]

    if deployment_id and (compute_cluster_id or nodepool_id):
      raise UserError(
          "You can only specify one of deployment_id or compute_cluster_id and nodepool_id.")

    runner_selector = None
    if deployment_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with deployment ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Deployment.get_runner_selector(
          user_id=user_id, deployment_id=deployment_id)
    elif compute_cluster_id and nodepool_id:
      if not user_id and not os.environ.get('CLARIFAI_USER_ID'):
        raise UserError(
            "User ID is required for model prediction with compute cluster ID and nodepool ID, please provide user_id in the method call."
        )
      if not user_id:
        user_id = os.environ.get('CLARIFAI_USER_ID')
      runner_selector = Nodepool.get_runner_selector(
          user_id=user_id, compute_cluster_id=compute_cluster_id, nodepool_id=nodepool_id)

    return self.stream(
        inputs=input_generator(),
        runner_selector=runner_selector,
        inference_params=inference_params,
        output_config=output_config)

  def _override_model_version(self, inference_params: Dict = {}, output_config: Dict = {}) -> None:
    """Overrides the model version.

    Args:
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
          min_value (float): The minimum value of the prediction confidence to filter.
          max_concepts (int): The maximum number of concepts to return.
          select_concepts (list[Concept]): The concepts to select.
          sample_ms (int): The number of milliseconds to sample.
    """
    params = Struct()
    if inference_params is not None:
      params.update(inference_params)

    self.model_info.model_version.output_info.CopyFrom(
        resources_pb2.OutputInfo(
            output_config=resources_pb2.OutputConfig(**output_config), params=params))

  def _list_concepts(self) -> List[str]:
    """Lists all the concepts for the model type.

    Returns:
        concepts (List): List of concepts for the model type.
    """
    request_data = dict(user_app_id=self.user_app_id)
    all_concepts_infos = self.list_pages_generator(self.STUB.ListConcepts,
                                                   service_pb2.ListConceptsRequest, request_data)
    return [concept_info['concept_id'] for concept_info in all_concepts_infos]

  def load_info(self) -> None:
    """Loads the model info."""
    request = service_pb2.GetModelRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        version_id=self.model_info.model_version.id)
    response = self._grpc_request(self.STUB.GetModel, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    dict_response = MessageToDict(response, preserving_proto_field_name=True)
    self.kwargs = self.process_response_keys(dict_response['model'])
    self.model_info = resources_pb2.Model(**self.kwargs)

  def __getattr__(self, name):
    return getattr(self.model_info, name)

  def __str__(self):
    if len(self.kwargs) < 10:
      self.load_info()

    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.model_info, param)}" for param in init_params
        if hasattr(self.model_info, param)
    ]
    return f"Model Details: \n{', '.join(attribute_strings)}\n"

  def list_evaluations(self) -> resources_pb2.EvalMetrics:
    """List all eval_metrics of current model version

    Raises:
        Exception: Failed to call API

    Returns:
        resources_pb2.EvalMetrics
    """
    assert self.model_info.model_version.id, "Model version is empty. Please provide `model_version` as arguments or with a URL as the format '{user_id}/{app_id}/models/{your_model_id}/model_version_id/{your_version_model_id}' when initializing."
    request = service_pb2.ListModelVersionEvaluationsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_version_id=self.model_info.model_version.id)
    response = self._grpc_request(self.STUB.ListModelVersionEvaluations, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    return response.eval_metrics

  def evaluate(self,
               dataset: Dataset = None,
               dataset_id: str = None,
               dataset_app_id: str = None,
               dataset_user_id: str = None,
               dataset_version_id: str = None,
               eval_id: str = None,
               extended_metrics: dict = None,
               eval_info: dict = None) -> resources_pb2.EvalMetrics:
    """ Run evaluation

    Args:
      dataset (Dataset): If Clarifai Dataset is set, it will ignore other arguments prefixed with 'dataset_'.
      dataset_id (str): Dataset Id. Default is None.
      dataset_app_id (str): App ID for cross app evaluation, leave it as None to use Model App ID. Default is None.
      dataset_user_id (str): User ID for cross app evaluation, leave it as None to use Model User ID. Default is None.
      dataset_version_id (str): Dataset version Id. Default is None.
      eval_id (str): Specific ID for the evaluation. You must specify this parameter to either overwrite the result with the dataset ID or format your evaluation in an informative manner. If you don't, it will use random ID from system. Default is None.
      extended_metrics (dict): user custom metrics result. Default is None.
      eval_info (dict): custom eval info. Default is empty dict.

    Return
      eval_metrics

    """
    assert self.model_info.model_version.id, "Model version is empty. Please provide `model_version` as arguments or with a URL as the format '{user_id}/{app_id}/models/{your_model_id}/model_version_id/{your_version_model_id}' when initializing."

    if dataset:
      self.logger.info("Using dataset, ignore other arguments prefixed with 'dataset_'")
      dataset_id = dataset.id
      dataset_app_id = dataset.app_id
      dataset_user_id = dataset.user_id
      dataset_version_id = dataset.version.id
    else:
      self.logger.warning(
          "Arguments prefixed with `dataset_` will be removed soon, please use dataset")

    gt_dataset = resources_pb2.Dataset(
        id=dataset_id,
        app_id=dataset_app_id or self.auth_helper.app_id,
        user_id=dataset_user_id or self.auth_helper.user_id,
        version=resources_pb2.DatasetVersion(id=dataset_version_id))

    metrics = None
    if isinstance(extended_metrics, dict):
      metrics = Struct()
      metrics.update(extended_metrics)
      metrics = resources_pb2.ExtendedMetrics(user_metrics=metrics)

    eval_info_params = None
    if isinstance(eval_info, dict):
      eval_info_params = Struct()
      eval_info_params.update(eval_info)
      eval_info_params = resources_pb2.EvalInfo(params=eval_info_params)

    eval_metric = resources_pb2.EvalMetrics(
        id=eval_id,
        model=resources_pb2.Model(
            id=self.id,
            app_id=self.auth_helper.app_id,
            user_id=self.auth_helper.user_id,
            model_version=resources_pb2.ModelVersion(id=self.model_info.model_version.id),
        ),
        extended_metrics=metrics,
        ground_truth_dataset=gt_dataset,
        eval_info=eval_info_params,
    )
    request = service_pb2.PostEvaluationsRequest(
        user_app_id=self.user_app_id,
        eval_metrics=[eval_metric],
    )
    response = self._grpc_request(self.STUB.PostEvaluations, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info(
        "\nModel evaluation in progress. Kindly allow a few minutes for completion. Processing time may vary based on the model and dataset sizes."
    )

    return response.eval_metrics

  def get_eval_by_id(
      self,
      eval_id: str,
      label_counts=False,
      test_set=False,
      binary_metrics=False,
      confusion_matrix=False,
      metrics_by_class=False,
      metrics_by_area=False,
  ) -> resources_pb2.EvalMetrics:
    """Get detail eval_metrics by eval_id with extra metric fields

    Args:
        eval_id (str): eval id
        label_counts (bool, optional): Set True to get label counts. Defaults to False.
        test_set (bool, optional): Set True to get test set. Defaults to False.
        binary_metrics (bool, optional): Set True to get binary metric. Defaults to False.
        confusion_matrix (bool, optional): Set True to get confusion matrix. Defaults to False.
        metrics_by_class (bool, optional): Set True to get metrics by class. Defaults to False.
        metrics_by_area (bool, optional): Set True to get metrics by area. Defaults to False.

    Raises:
        Exception: Failed to call API

    Returns:
        resources_pb2.EvalMetrics: eval_metrics
    """
    request = service_pb2.GetEvaluationRequest(
        user_app_id=self.user_app_id,
        evaluation_id=eval_id,
        fields=resources_pb2.FieldsValue(
            label_counts=label_counts,
            test_set=test_set,
            binary_metrics=binary_metrics,
            confusion_matrix=confusion_matrix,
            metrics_by_class=metrics_by_class,
            metrics_by_area=metrics_by_area,
        ))
    response = self._grpc_request(self.STUB.GetEvaluation, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)

    return response.eval_metrics

  def get_latest_eval(self,
                      label_counts=False,
                      test_set=False,
                      binary_metrics=False,
                      confusion_matrix=False,
                      metrics_by_class=False,
                      metrics_by_area=False) -> Union[resources_pb2.EvalMetrics, None]:
    """
    Run `get_eval_by_id` method with latest `eval_id`

    Args:
      label_counts (bool, optional): Set True to get label counts. Defaults to False.
      test_set (bool, optional): Set True to get test set. Defaults to False.
      binary_metrics (bool, optional): Set True to get binary metric. Defaults to False.
      confusion_matrix (bool, optional): Set True to get confusion matrix. Defaults to False.
      metrics_by_class (bool, optional): Set True to get metrics by class. Defaults to False.
      metrics_by_area (bool, optional): Set True to get metrics by area. Defaults to False.

    Returns:
      eval_metric if model is evaluated otherwise None.

    """

    _latest = self.list_evaluations()[0]
    result = None
    if _latest.status.code == status_code_pb2.MODEL_EVALUATED:
      result = self.get_eval_by_id(
          eval_id=_latest.id,
          label_counts=label_counts,
          test_set=test_set,
          binary_metrics=binary_metrics,
          confusion_matrix=confusion_matrix,
          metrics_by_class=metrics_by_class,
          metrics_by_area=metrics_by_area)

    return result

  def get_eval_by_dataset(self, dataset: Dataset) -> List[resources_pb2.EvalMetrics]:
    """Get all eval data of dataset

    Args:
        dataset (Dataset): Clarifai dataset

    Returns:
        List[resources_pb2.EvalMetrics]
    """
    _id = dataset.id
    app = dataset.app_id or self.app_id
    user_id = dataset.user_id or self.user_id
    version = dataset.version.id

    list_eval: resources_pb2.EvalMetrics = self.list_evaluations()
    outputs = []
    for _eval in list_eval:
      if _eval.status.code == status_code_pb2.MODEL_EVALUATED:
        gt_ds = _eval.ground_truth_dataset
        if (_id == gt_ds.id and user_id == gt_ds.user_id and app == gt_ds.app_id):
          if not version or version == gt_ds.version.id:
            outputs.append(_eval)

    return outputs

  def get_raw_eval(self,
                   dataset: Dataset = None,
                   eval_id: str = None,
                   return_format: str = 'array') -> Union[resources_pb2.EvalTestSetEntry, Tuple[
                       np.array, np.array, list, List[Input]], Tuple[List[dict], List[dict]]]:
    """Get ground truths, predictions and input information. Do not pass dataset and eval_id at same time

    Args:
        dataset (Dataset): Clarifai dataset, get eval data of latest eval result of dataset.
        eval_id (str): Evaluation ID, get eval data of specific eval id.
        return_format (str, optional): Choice {proto, array, coco}. !Note that `coco` is only applicable for 'visual-detector'. Defaults to 'array'.

    Returns:

        Depends on `return_format`.

        * if return_format == proto
          `resources_pb2.EvalTestSetEntry`

        * if return_format == array
          `Tuple(np.array, np.array, List[str], List[Input])`: Tuple has 4 elements (y, y_pred, concept_ids, inputs).
            y, y_pred, concept_ids can be used to compute metrics. 'inputs' can be use to download
            - if model is 'classifier': 'y' and 'y_pred' are both arrays with a shape of (num_inputs,)
            - if model is 'visual-detector': 'y' and 'y_pred' are arrays with a shape of (num_inputs,), where each element is array has shape (num_annotation, 6) consists of [x_min, y_min, x_max, y_max, concept_index, score]. The score is always 1 for 'y'

        * if return_format == coco: Applicable only for 'visual-detector'
          `Tuple[List[Dict], List[Dict]]`: Tuple has 2 elemnts where first element is COCO Ground Truth and last one is COCO Prediction Annotation

    Example Usages:
    ------
    * Evaluate `visual-classifier` using sklearn

    ```python
    import os
    from sklearn.metrics import accuracy_score
    from sklearn.metrics import classification_report
    import numpy as np
    from clarifai.client.model import Model
    from clarifai.client.dataset import Dataset
    os.environ["CLARIFAI_PAT"] = "???"
    model = Model(url="url/of/model/includes/version-id")
    dataset = Dataset(dataset_id="dataset-id")
    y, y_pred, clss, input_protos = model.get_raw_eval(dataset, return_format="array")
    y = np.argmax(y, axis=1)
    y_pred = np.argmax(y_pred, axis=1)
    report = classification_report(y, y_pred, target_names=clss)
    print(report)
    acc = accuracy_score(y, y_pred)
    print("acc ", acc)
    ```

    * Evaluate `visual-detector` using COCOeval

    ```python
    import os
    import json
    from pycocotools.coco import COCO
    from pycocotools.cocoeval import COCOeval
    from clarifai.client.model import Model
    from clarifai.client.dataset import Dataset
    os.environ["CLARIFAI_PAT"] = "???" # Insert your PAT
    model = Model(url=model_url)
    dataset = Dataset(url=dataset_url)
    y, y_pred = model.get_raw_eval(dataset, return_format="coco")
    # save as files to load in COCO API
    def save_annot(d, path):
      with open(path, "w") as fp:
        json.dump(d, fp, indent=2)
    gt_path = os.path.join("gt.json")
    pred_path = os.path.join("pred.json")
    save_annot(y, gt_path)
    save_annot(y_pred, pred_path)

    cocoGt = COCO(gt_path)
    cocoPred = COCO(pred_path)
    cocoEval = COCOeval(cocoGt, cocoPred, "bbox")
    cocoEval.evaluate()
    cocoEval.accumulate()
    cocoEval.summarize() # Print out result of all classes with all area type
    # Example:
    # Average Precision  (AP) @[ IoU=0.50:0.95 | area=   all | maxDets=100 ] = 0.863
    # Average Precision  (AP) @[ IoU=0.50      | area=   all | maxDets=100 ] = 0.973
    # Average Precision  (AP) @[ IoU=0.75      | area=   all | maxDets=100 ] = 0.939
    # ...
    ```

    """
    from clarifai.utils.evaluation.testset_annotation_parser import (
        parse_eval_annotation_classifier, parse_eval_annotation_detector,
        parse_eval_annotation_detector_coco)

    valid_model_types = ["visual-classifier", "text-classifier", "visual-detector"]
    supported_format = ['proto', 'array', 'coco']
    assert return_format in supported_format, ValueError(
        f"Expected return_format in {supported_format}, got {return_format}")
    self.load_info()
    model_type_id = self.model_info.model_type_id
    assert model_type_id in valid_model_types, \
      f"This method only supports model types {valid_model_types}, but your model type is {self.model_info.model_type_id}."
    assert not (dataset and
                eval_id), "Using both `dataset` and `eval_id`, but only one should be passed."
    assert not dataset or not eval_id, "Please provide either `dataset` or `eval_id`, but nothing was passed."
    if model_type_id.endswith("-classifier") and return_format == "coco":
      raise ValueError(
          f"return_format coco only applies for `visual-detector`, however your model is `{model_type_id}`"
      )

    if dataset:
      eval_by_ds = self.get_eval_by_dataset(dataset)
      if len(eval_by_ds) == 0:
        raise Exception(f"Model is not valuated with dataset: {dataset}")
      eval_id = eval_by_ds[0].id

    detail_eval_data = self.get_eval_by_id(eval_id=eval_id, test_set=True, metrics_by_class=True)

    if return_format == "proto":
      return detail_eval_data.test_set
    else:
      if model_type_id.endswith("-classifier"):
        return parse_eval_annotation_classifier(detail_eval_data)
      elif model_type_id == "visual-detector":
        if return_format == "array":
          return parse_eval_annotation_detector(detail_eval_data)
        elif return_format == "coco":
          return parse_eval_annotation_detector_coco(detail_eval_data)

  def export(self, export_dir: str = None) -> None:
    """Export the model, stores the exported model as model.tar file

    Args:
        export_dir (str, optional): If provided, the exported model will be saved in the specified directory else export status will be shown. Defaults to None.

    Example:
        >>> from clarifai.client.model import Model
        >>> model = Model("url")
        >>> model.export()
                or
        >>> model.export('/path/to/export_model_dir')
    """
    assert self.model_info.model_version.id, "Model version ID is missing. Please provide a `model_version` with a valid `id` as an argument or as a URL in the following format: '{user_id}/{app_id}/models/{your_model_id}/model_version_id/{your_version_model_id}' when initializing."
    if export_dir:
      try:
        if not os.path.exists(export_dir):
          os.makedirs(export_dir)
      except OSError as e:
        raise Exception(f"An error occurred while creating the directory: {e}")

    def _get_export_response():
      get_export_request = service_pb2.GetModelVersionExportRequest(
          user_app_id=self.user_app_id,
          model_id=self.id,
          version_id=self.model_info.model_version.id,
      )
      response = self._grpc_request(self.STUB.GetModelVersionExport, get_export_request)

      if response.status.code != status_code_pb2.SUCCESS and response.status.code != status_code_pb2.CONN_DOES_NOT_EXIST:
        raise Exception(response.status)

      return response

    def _download_exported_model(
        get_model_export_response: service_pb2.SingleModelVersionExportResponse,
        local_filepath: str):
      model_export_url = get_model_export_response.export.url
      model_export_file_size = get_model_export_response.export.size

      with open(local_filepath, 'wb') as f:
        progress = tqdm(
            total=model_export_file_size, unit='B', unit_scale=True, desc="Exporting model")
        downloaded_size = 0
        range_size = RANGE_SIZE
        chunk_size = CHUNK_SIZE
        retry = False
        retry_count = 0
        while downloaded_size < model_export_file_size:
          if downloaded_size + range_size >= model_export_file_size:
            range_header = f"bytes={downloaded_size}-"
          else:
            range_header = f"bytes={downloaded_size}-{(downloaded_size+range_size-1)}"
          try:
            session = requests.Session()
            retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
            session.mount('https://', HTTPAdapter(max_retries=retries))
            session.headers.update({'Authorization': self.metadata[0][1], 'Range': range_header})
            response = session.get(model_export_url, stream=True)
            response.raise_for_status()

            for chunk in response.iter_content(chunk_size=chunk_size):
              f.write(chunk)
              progress.update(len(chunk))
            f.flush()
            os.fsync(f.fileno())
            downloaded_size += range_size
            if not retry:
              range_size = (
                  range_size * 2) if (range_size * 2) < MAX_RANGE_SIZE else MAX_RANGE_SIZE
              chunk_size = (
                  chunk_size * 2) if (chunk_size * 2) < MAX_CHUNK_SIZE else MAX_CHUNK_SIZE
          except Exception as e:
            self.logger.error(f"Error downloading model: {e}")
            range_size = (
                range_size // 2) if (range_size // 2) > MIN_RANGE_SIZE else MIN_RANGE_SIZE
            chunk_size = (
                chunk_size // 2) if (chunk_size // 2) > MIN_CHUNK_SIZE else MIN_CHUNK_SIZE
            retry = True
            retry_count += 1
            f.seek(downloaded_size)
            progress.reset(total=model_export_file_size)
            progress.update(downloaded_size)
            if retry_count > 5:
              break
      progress.close()

      self.logger.info(
          f"Model ID {self.id} with version {self.model_info.model_version.id} exported successfully to {export_dir}/model.tar"
      )

    get_export_response = _get_export_response()
    if get_export_response.status.code == status_code_pb2.CONN_DOES_NOT_EXIST:
      put_export_request = service_pb2.PutModelVersionExportsRequest(
          user_app_id=self.user_app_id,
          model_id=self.id,
          version_id=self.model_info.model_version.id,
      )

      response = self._grpc_request(self.STUB.PutModelVersionExports, put_export_request)
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(response.status)

      self.logger.info(
          f"Export process has started for Model ID {self.id}, Version {self.model_info.model_version.id}"
      )
      if export_dir:
        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        while True:
          get_export_response = _get_export_response()
          if (get_export_response.export.status.code == status_code_pb2.MODEL_EXPORTING or \
                get_export_response.export.status.code == status_code_pb2.MODEL_EXPORT_PENDING) and \
                  time.time() - start_time < MODEL_EXPORT_TIMEOUT:
            self.logger.info(
                f"Export process is ongoing for Model ID {self.id}, Version {self.model_info.model_version.id}. Please wait..."
            )
            time.sleep(next(backoff_iterator))
          elif get_export_response.export.status.code == status_code_pb2.MODEL_EXPORTED:
            _download_exported_model(get_export_response, os.path.join(export_dir, "model.tar"))
            break
          elif time.time() - start_time > MODEL_EXPORT_TIMEOUT:
            raise Exception(
                f"""Model Export took too long. Please try again or contact support@clarifai.com
                Req ID: {get_export_response.status.req_id}""")
    elif get_export_response.export.status.code == status_code_pb2.MODEL_EXPORTED:
      if export_dir:
        _download_exported_model(get_export_response, os.path.join(export_dir, "model.tar"))
      else:
        self.logger.info(
            f"Model ID {self.id} with version {self.model_info.model_version.id} is already exported, you can download it from the following URL: {get_export_response.export.url}"
        )
    elif get_export_response.export.status.code == status_code_pb2.MODEL_EXPORTING or \
            get_export_response.export.status.code == status_code_pb2.MODEL_EXPORT_PENDING:
      self.logger.info(
          f"Export process is ongoing for Model ID {self.id}, Version {self.model_info.model_version.id}. Please wait..."
      )

  @staticmethod
  def _make_pretrained_config_proto(input_field_maps: dict,
                                    output_field_maps: dict,
                                    url: str = None):
    """Make PretrainedModelConfig for uploading new version

    Args:
        input_field_maps (dict): dict
        output_field_maps (dict): dict
        url (str, optional): direct download url. Defaults to None.
    """

    def _parse_fields_map(x):
      """parse input, outputs to Struct"""
      _fields_map = Struct()
      _fields_map.update(x)
      return _fields_map

    input_fields_map = _parse_fields_map(input_field_maps)
    output_fields_map = _parse_fields_map(output_field_maps)

    return resources_pb2.PretrainedModelConfig(
        input_fields_map=input_fields_map, output_fields_map=output_fields_map, model_zip_url=url)

  @staticmethod
  def _make_inference_params_proto(
      inference_parameters: List[Dict]) -> List[resources_pb2.ModelTypeField]:
    """Convert list of Clarifai inference parameters to proto for uploading new version

    Args:
        inference_parameters (List[Dict]): Each dict has keys {field_type, path, default_value, description}

    Returns:
        List[resources_pb2.ModelTypeField]
    """

    def _make_default_value_proto(dtype, value):
      if dtype == 1:
        return Value(bool_value=value)
      elif dtype == 2 or dtype == 21:
        return Value(string_value=value)
      elif dtype == 3:
        return Value(number_value=value)

    iterative_proto_params = []
    for param in inference_parameters:
      dtype = param.get("field_type")
      proto_param = resources_pb2.ModelTypeField(
          path=param.get("path"),
          field_type=dtype,
          default_value=_make_default_value_proto(dtype=dtype, value=param.get("default_value")),
          description=param.get("description"),
      )
      iterative_proto_params.append(proto_param)
    return iterative_proto_params

  def create_version_by_file(self,
                             file_path: str,
                             input_field_maps: dict,
                             output_field_maps: dict,
                             inference_parameter_configs: dict = None,
                             model_version: str = None,
                             part_id: int = 1,
                             range_start: int = 0,
                             no_cache: bool = False,
                             no_resume: bool = False,
                             description: str = "") -> 'Model':
    """Create model version by uploading local file

    Args:
        file_path (str): path to built file.
        input_field_maps (dict): a dict where the key is clarifai input field and the value is triton model input,
          {clarifai_input_field: triton_input_filed}.
        output_field_maps (dict): a dict where the keys are clarifai output fields and the values are triton model outputs,
          {clarifai_output_field1: triton_output_filed1, clarifai_output_field2: triton_output_filed2,...}.
        inference_parameter_configs (List[dict]): list of dicts - keys are path, field_type, default_value, description. Default is None
        model_version (str, optional): Custom model version. Defaults to None.
        part_id (int, optional): part id of file. Defaults to 1.
        range_start (int, optional): range of uploaded size. Defaults to 0.
        no_cache (bool, optional): not saving uploading cache that is used to resume uploading. Defaults to False.
        no_resume (bool, optional): disable auto resume upload. Defaults to False.
        description (str): Model description.

    Return:
      Model: instance of Model with new created version

    """
    file_size = os.path.getsize(file_path)
    assert MIN_CHUNK_FOR_UPLOAD_FILE <= file_size <= MAX_CHUNK_FOR_UPLOAD_FILE, "The file size exceeds the allowable limit, which ranges from 5MiB to 5GiB."

    pretrained_proto = Model._make_pretrained_config_proto(
        input_field_maps=input_field_maps, output_field_maps=output_field_maps)
    inference_param_proto = Model._make_inference_params_proto(
        inference_parameter_configs) if inference_parameter_configs else None

    if file_size >= 1e9:
      chunk_size = 1024 * 50_000  # 50MB
    else:
      chunk_size = 1024 * 10_000  # 10MB

    #self.logger.info(f"Chunk {chunk_size/1e6}MB, {file_size/chunk_size} steps")
    #self.logger.info(f" Max bytes per stream {MAX_SIZE_PER_STREAM}")

    cache_dir = os.path.join(file_path, '..', '.cache')
    cache_upload_file = os.path.join(cache_dir, "upload.json")
    last_percent = 0
    if os.path.exists(cache_upload_file) and not no_resume:
      with open(cache_upload_file, "r") as fp:
        try:
          cache_info = json.load(fp)
          if isinstance(cache_info, dict):
            part_id = cache_info.get("part_id", part_id)
            chunk_size = cache_info.get("chunk_size", chunk_size)
            range_start = cache_info.get("range_start", range_start)
            model_version = cache_info.get("model_version", model_version)
            last_percent = cache_info.get("last_percent", last_percent)
        except Exception as e:
          self.logger.error(f"Skipping loading the upload cache due to error {e}.")

    def init_model_version_upload(model_version):
      return service_pb2.PostModelVersionsUploadRequest(
          upload_config=service_pb2.PostModelVersionsUploadConfig(
              user_app_id=self.user_app_id,
              model_id=self.id,
              total_size=file_size,
              model_version=resources_pb2.ModelVersion(
                  id=model_version,
                  pretrained_model_config=pretrained_proto,
                  description=description,
                  output_info=resources_pb2.OutputInfo(params_specs=inference_param_proto)),
          ))

    def _uploading(chunk, part_id, range_start, model_version):
      return service_pb2.PostModelVersionsUploadRequest(
          content_part=resources_pb2.UploadContentPart(
              data=chunk, part_number=part_id, range_start=range_start))

    finished_status = [status_code_pb2.SUCCESS, status_code_pb2.UPLOAD_DONE]
    uploading_in_progress_status = [
        status_code_pb2.UPLOAD_IN_PROGRESS, status_code_pb2.MODEL_UPLOADING
    ]

    def _save_cache(cache: dict):
      if not no_cache:
        os.makedirs(cache_dir, exist_ok=True)
        with open(cache_upload_file, "w") as fp:
          json.dump(cache, fp, indent=2)

    def stream_request(fp, part_id, end_part_id, chunk_size, version):
      yield init_model_version_upload(version)
      for iter_part_id in range(part_id, end_part_id):
        chunk = fp.read(chunk_size)
        if not chunk:
          return
        yield _uploading(
            chunk=chunk,
            part_id=iter_part_id,
            range_start=chunk_size * (iter_part_id - 1),
            model_version=version)

    tqdm_loader = tqdm(total=100)
    if model_version:
      desc = f"Uploading model `{self.id}` version `{model_version}` ..."
    else:
      desc = f"Uploading model `{self.id}` ..."
    tqdm_loader.set_description(desc)

    cache_uploading_info = {}
    cache_uploading_info["part_id"] = part_id
    cache_uploading_info["model_version"] = model_version
    cache_uploading_info["range_start"] = range_start
    cache_uploading_info["chunk_size"] = chunk_size
    cache_uploading_info["last_percent"] = last_percent
    tqdm_loader.update(last_percent)
    last_part_id = part_id
    n_chunks = file_size // chunk_size
    n_chunk_per_stream = MAX_SIZE_PER_STREAM // chunk_size or 1

    def stream_and_logging(request, tqdm_loader, cache_uploading_info, expected_steps: int = None):
      for st_step, st_response in enumerate(self.auth_helper.get_stub().PostModelVersionsUpload(
          request, metadata=self.auth_helper.metadata)):
        if st_response.status.code in uploading_in_progress_status:
          if cache_uploading_info["model_version"]:
            assert st_response.model_version_id == cache_uploading_info[
                "model_version"], RuntimeError
          else:
            cache_uploading_info["model_version"] = st_response.model_version_id
          if st_step > 0:
            cache_uploading_info["part_id"] += 1
            cache_uploading_info["range_start"] += chunk_size
            _save_cache(cache_uploading_info)

            if st_response.status.percent_completed:
              step_percent = st_response.status.percent_completed - cache_uploading_info["last_percent"]
              cache_uploading_info["last_percent"] += step_percent
              tqdm_loader.set_description(
                  f"{st_response.status.description}, {st_response.status.details}, version id  {cache_uploading_info.get('model_version')}"
              )
              tqdm_loader.update(step_percent)
        elif st_response.status.code not in finished_status + uploading_in_progress_status:
          # TODO: Find better way to handle error
          if expected_steps and st_step < expected_steps:
            raise Exception(f"Failed to upload model, error: {st_response.status}")

    with open(file_path, 'rb') as fp:
      # seeking
      for _ in range(1, last_part_id):
        fp.read(chunk_size)
      # Stream even part
      end_part_id = n_chunks or 1
      for iter_part_id in range(int(last_part_id), int(n_chunks), int(n_chunk_per_stream)):
        end_part_id = iter_part_id + n_chunk_per_stream
        if end_part_id >= n_chunks:
          end_part_id = n_chunks
        expected_steps = end_part_id - iter_part_id + 1  # init step
        st_reqs = stream_request(
            fp,
            iter_part_id,
            end_part_id=end_part_id,
            chunk_size=chunk_size,
            version=cache_uploading_info["model_version"])
        stream_and_logging(st_reqs, tqdm_loader, cache_uploading_info, expected_steps)
      # Stream last part
      accum_size = (end_part_id - 1) * chunk_size
      remained_size = file_size - accum_size if accum_size >= 0 else file_size
      st_reqs = stream_request(
          fp,
          end_part_id,
          end_part_id=end_part_id + 1,
          chunk_size=remained_size,
          version=cache_uploading_info["model_version"])
      stream_and_logging(st_reqs, tqdm_loader, cache_uploading_info, 2)

    # clean up cache
    if not no_cache:
      try:
        os.remove(cache_upload_file)
      except Exception:
        _save_cache({})

    if cache_uploading_info["last_percent"] <= 100:
      tqdm_loader.update(100 - cache_uploading_info["last_percent"])
      tqdm_loader.set_description("Upload done")

    tqdm_loader.set_description(
        f"Success uploading model {self.id}, new version {cache_uploading_info.get('model_version')}"
    )

    return Model.from_auth_helper(
        auth=self.auth_helper,
        model_id=self.id,
        model_version=dict(id=cache_uploading_info.get('model_version')))

  def create_version_by_url(self,
                            url: str,
                            input_field_maps: dict,
                            output_field_maps: dict,
                            inference_parameter_configs: List[dict] = None,
                            description: str = "") -> 'Model':
    """Upload a new version of an existing model in the Clarifai platform using direct download url.

    Args:
      url (str]): url of zip of model
      input_field_maps (dict): a dict where the key is clarifai input field and the value is triton model input,
          {clarifai_input_field: triton_input_filed}.
      output_field_maps (dict): a dict where the keys are clarifai output fields and the values are triton model outputs,
          {clarifai_output_field1: triton_output_filed1, clarifai_output_field2: triton_output_filed2,...}.
      inference_parameter_configs (List[dict]): list of dicts - keys are path, field_type, default_value, description. Default is None
      description (str): Model description.

    Return:
      Model: instance of Model with new created version
    """

    pretrained_proto = Model._make_pretrained_config_proto(
        input_field_maps=input_field_maps, output_field_maps=output_field_maps, url=url)
    inference_param_proto = Model._make_inference_params_proto(
        inference_parameter_configs) if inference_parameter_configs else None
    request = service_pb2.PostModelVersionsRequest(
        user_app_id=self.user_app_id,
        model_id=self.id,
        model_versions=[
            resources_pb2.ModelVersion(
                pretrained_model_config=pretrained_proto,
                description=description,
                output_info=resources_pb2.OutputInfo(params_specs=inference_param_proto))
        ])
    response = self._grpc_request(self.STUB.PostModelVersions, request)

    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(f"Failed to upload model, error: {response.status}")
    self.logger.info(
        f"Success uploading model {self.id}, new version {response.model.model_version.id}")

    return Model.from_auth_helper(
        auth=self.auth_helper,
        model_id=self.id,
        model_version=dict(id=response.model.model_version.id))



from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Type

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct

from clarifai.client.input import Inputs
from clarifai.utils.misc import get_uuid

from .base import ClarifaiDataLoader, ClarifaiDataset


class TextClassificationDataset(ClarifaiDataset):
  """Upload text classification datasets to clarifai datasets"""

  def __init__(self,
               data_generator: Type[ClarifaiDataLoader],
               dataset_id: str,
               max_workers: int = 4) -> None:
    super().__init__(data_generator, dataset_id, max_workers)

  def _extract_protos(self, batch_input_ids: List[int]
                     ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
    """Create input image and annotation protos for batch of input ids.
    Args:
      batch_input_ids: List of input IDs to retrieve the protos for.
    Returns:
      input_protos: List of input protos.
      annotation_protos: List of annotation protos.
    """
    input_protos, annotation_protos = [], []

    def process_data_item(id):
      data_item = self.data_generator[id]
      metadata = Struct()
      text = data_item.text
      labels = data_item.labels if ((data_item.labels is None) or
                                    isinstance(data_item.labels, list)) else [data_item.labels
                                                                             ]  # clarifai concept
      label_ids = data_item.label_ids
      input_id = f"{self.dataset_id}-{get_uuid(8)}" if data_item.id is None else f"{self.dataset_id}-{str(data_item.id)}"
      if data_item.metadata is not None:
        metadata.update(data_item.metadata)

      self.all_input_ids[id] = input_id
      input_protos.append(
          Inputs.get_text_input(
              input_id=input_id,
              raw_text=text,
              dataset_id=self.dataset_id,
              labels=labels,
              label_ids=label_ids,
              metadata=metadata))

    with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
      futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
      for job in futures:
        job.result()

    return input_protos, annotation_protos



from itertools import tee
from typing import Iterator

from clarifai_grpc.grpc.api import service_pb2, service_pb2_grpc
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2

from ..utils.url_fetcher import ensure_urls_downloaded


class ModelServicer(service_pb2_grpc.V2Servicer):
  """
  This is the servicer that will handle the gRPC requests from either the dev server or runner loop.
  """

  def __init__(self, model):
    """
    Args:
        model: The class that will handle the model logic. Must implement predict(),
    generate(), stream().
    """
    self.model = model

  def PostModelOutputs(self, request: service_pb2.PostModelOutputsRequest,
                       context=None) -> service_pb2.MultiOutputResponse:
    """
    This is the method that will be called when the servicer is run. It takes in an input and
    returns an output.
    """

    # Download any urls that are not already bytes.
    ensure_urls_downloaded(request)

    try:
      return self.model.predict_wrapper(request)
    except Exception as e:
      return service_pb2.MultiOutputResponse(status=status_pb2.Status(
          code=status_code_pb2.MODEL_PREDICTION_FAILED,
          description="Failed",
          details="",
          internal_details=str(e),
      ))

  def GenerateModelOutputs(self, request: service_pb2.PostModelOutputsRequest,
                           context=None) -> Iterator[service_pb2.MultiOutputResponse]:
    """
    This is the method that will be called when the servicer is run. It takes in an input and
    returns an output.
    """
    # Download any urls that are not already bytes.
    ensure_urls_downloaded(request)

    try:
      return self.model.generate_wrapper(request)
    except Exception as e:
      yield service_pb2.MultiOutputResponse(status=status_pb2.Status(
          code=status_code_pb2.MODEL_PREDICTION_FAILED,
          description="Failed",
          details="",
          internal_details=str(e),
      ))

  def StreamModelOutputs(self,
                         request: Iterator[service_pb2.PostModelOutputsRequest],
                         context=None) -> Iterator[service_pb2.MultiOutputResponse]:
    """
    This is the method that will be called when the servicer is run. It takes in an input and
    returns an output.
    """
    # Duplicate the iterator
    request, request_copy = tee(request)

    # Download any urls that are not already bytes.
    for req in request:
      ensure_urls_downloaded(req)

    try:
      return self.model.stream_wrapper(request_copy)
    except Exception as e:
      yield service_pb2.MultiOutputResponse(status=status_pb2.Status(
          code=status_code_pb2.MODEL_PREDICTION_FAILED,
          description="Failed",
          details="",
          internal_details=str(e),
      ))



import os

registry = os.environ.get('CLARIFAI_BASE_IMAGE_REGISTRY', 'public.ecr.aws/clarifai-models')

GIT_SHA = "df565436eea93efb3e8d1eb558a0a46df29523ec"

PYTHON_BASE_IMAGE = registry + '/python-base:{python_version}-' + GIT_SHA
TORCH_BASE_IMAGE = registry + '/torch:{torch_version}-py{python_version}-cuda{cuda_version}-' + GIT_SHA

# List of available python base images
AVAILABLE_PYTHON_IMAGES = ['3.11', '3.12']

DEFAULT_PYTHON_VERSION = 3.12

# List of available torch images
# Keep sorted by most recent cuda version.
AVAILABLE_TORCH_IMAGES = [
    '2.4.1-py3.11-cuda124',
    '2.5.1-py3.11-cuda124',
    '2.4.1-py3.12-cuda124',
    '2.5.1-py3.12-cuda124',
    # '2.4.1-py3.13-cuda124',
    # '2.5.1-py3.13-cuda124',
]
CONCEPTS_REQUIRED_MODEL_TYPE = [
    'visual-classifier', 'visual-detector', 'visual-segmenter', 'text-classifier'
]
